#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
sha2: A non-optimized but very compact implementation of the SHA2 hash
      algorithm for the MMGen suite.  Implements SHA256, SHA512 and
      SHA256Compress (unpadded SHA256, required for Zcash addresses)
"""

# IMPORTANT NOTE: Since GMP precision is platform-dependent, generated constants
# for SHA512 are not guaranteed to be correct!  Therefore, the SHA512
# implementation must not be used for anything but testing and study.  Test with
# the test/hashfunc.py script in the MMGen repository.

from struct import pack, unpack

class Sha2:
	'Implementation based on the pseudocode at https://en.wikipedia.org/wiki/SHA-2'
	K = ()

	@classmethod
	def initConstants(cls):

		from math import sqrt

		def nextPrime(n=2):
			while True:
				for factor in range(2, int(sqrt(n))+1):
					if n % factor == 0:
						break
				else:
					yield n
				n += 1

		# Find the first nRounds primes
		npgen = nextPrime()
		primes = [next(npgen) for x in range(cls.nRounds)]

		fb_mul = 2 ** cls.wordBits
		def getFractionalBits(n):
			return int((n - int(n)) * fb_mul)

		if cls.use_gmp:
			from gmpy2 import context, set_context, sqrt, cbrt
			# context() parameters are platform-dependent!
			set_context(context(precision=75, round=1)) # OK for gmp 6.1.2 / gmpy 2.1.0
		else:
			cbrt = lambda n: pow(n, 1 / 3)

		# First wordBits bits of the fractional parts of the square roots of the first 8 primes
		cls.H_init = tuple(getFractionalBits(sqrt(n)) for n in primes[:8])

		# First wordBits bits of the fractional parts of the cube roots of the first nRounds primes
		cls.K = tuple(getFractionalBits(cbrt(n)) for n in primes)

	def __init__(self, message, *, preprocess=True):
		'Use preprocess=False for Sha256Compress'
		assert isinstance(message, (bytes, bytearray, list)), 'message must be of type bytes, bytearray or list'
		if not self.K:
			type(self).initConstants()
		self.H = list(self.H_init)
		self.M = message
		self.W = [0] * self.nRounds
		if preprocess:
			self.padMessage()
		self.bytesToWords()
		self.compute()

	def padMessage(self):
		"""
		Pre-processing (padding) (adapted from the standard, translating bits to bytes)
		- Begin with the original message of length MSGLEN bytes
		- Create MLPACK := MSGLEN * 8 encoded as a blkSize-bit big-endian integer
		- Append 0x80 (0b10000000)
		- Append PADLEN null bytes, where PADLEN is the minimum number >= 0 such that
		  MSGLEN + 1 + PADLEN + len(MLPACK) is a multiple of blkSize
		- Append MLPACK
		"""
		mlpack = self.pack_msglen()
		padlen = self.blkSize - (len(self.M) % self.blkSize) - len(mlpack) - 1
		if padlen < 0:
			padlen += self.blkSize
		self.M = self.M + bytes([0x80] + [0] * padlen) + mlpack

	def bytesToWords(self):
		ws = self.wordSize
		assert len(self.M) % ws == 0
		self.M = tuple(unpack(self.word_fmt, self.M[i*ws:ws+(i*ws)])[0] for i in range(len(self.M) // ws))

	def digest(self):
		return b''.join((pack(self.word_fmt, w) for w in self.H))

	def hexdigest(self):
		return self.digest().hex()

	def compute(self):
		for i in range(0, len(self.M), 16):
			self.processBlock(i)

	def processBlock(self, offset):
		'Process a blkSize-byte chunk of the message'

		def rrotate(a, b):
			return ((a << self.wordBits-b) & self.wordMask) | (a >> b)

		def addm(a, b):
			return (a + b) & self.wordMask

		# Copy chunk into first 16 words of message schedule array
		for i in range(16):
			self.W[i] = self.M[offset + i]

		# Extend the first 16 words into the remaining nRounds words of message schedule array
		for i in range(16, self.nRounds):
			g0 = self.W[i-15]
			gamma0 = rrotate(g0, self.g0r1) ^ rrotate(g0, self.g0r2) ^ (g0 >> self.g0r3)
			g1 = self.W[i-2]
			gamma1 = rrotate(g1, self.g1r1) ^ rrotate(g1, self.g1r2) ^ (g1 >> self.g1r3)
			self.W[i] = addm(addm(addm(gamma0, self.W[i-7]), gamma1), self.W[i-16])

		# Initialize working variables from current hash state
		a, b, c, d, e, f, g, h = self.H

		# Compression function main loop
		for i in range(self.nRounds):
			ch = (e & f) ^ (~e & g)
			maj = (a & b) ^ (a & c) ^ (b & c)

			sigma0 = rrotate(a, self.s0r1) ^ rrotate(a, self.s0r2) ^ rrotate(a, self.s0r3)
			sigma1 = rrotate(e, self.s1r1) ^ rrotate(e, self.s1r2) ^ rrotate(e, self.s1r3)

			t1 = addm(addm(addm(addm(h, sigma1), ch), self.K[i]), self.W[i])
			t2 = addm(sigma0, maj)

			h = g
			g = f
			f = e
			e = addm(d, t1)
			d = c
			c = b
			b = a
			a = addm(t1, t2)

		# Save hash state
		for n, v in enumerate((a, b, c, d, e, f, g, h)):
			self.H[n] = addm(self.H[n], v)

class Sha256(Sha2):
	use_gmp = False
	g0r1, g0r2, g0r3 = (7, 18, 3)
	g1r1, g1r2, g1r3 = (17, 19, 10)
	s0r1, s0r2, s0r3 = (2, 13, 22)
	s1r1, s1r2, s1r3 = (6, 11, 25)
	blkSize = 64
	nRounds = 64
	wordSize = 4
	wordBits = 32
	wordMask = 0xffffffff
	word_fmt = '>I'

	def pack_msglen(self):
		return pack('>Q', len(self.M)*8)

class Sha512(Sha2):
	"""
	SHA-512 is identical in structure to SHA-256, but:
	- the message is broken into 1024-bit chunks,
	- the initial hash values and round constants are extended to 64 bits,
	- there are 80 rounds instead of 64,
	- the message schedule array W has 80 64-bit words instead of 64 32-bit words,
	- to extend the message schedule array W, the loop is from 16 to 79 instead of from 16 to 63,
	- the round constants are based on the first 80 primes 2..409,
	- the word size used for calculations is 64 bits long,
	- the appended length of the message (before pre-processing), in bits, is a 128-bit big-endian integer, and
	- the shift and rotate amounts used are different.
	"""
	use_gmp = True
	g0r1, g0r2, g0r3 = (1, 8, 7)
	g1r1, g1r2, g1r3 = (19, 61, 6)
	s0r1, s0r2, s0r3 = (28, 34, 39)
	s1r1, s1r2, s1r3 = (14, 18, 41)
	blkSize = 128
	nRounds = 80
	wordSize = 8
	wordBits = 64
	wordMask = 0xffffffffffffffff
	word_fmt = '>Q'

	def pack_msglen(self):
		return  pack('>Q', (len(self.M)*8) // (2**64)) + \
				pack('>Q', (len(self.M)*8) % (2**64))



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
test/clean.py: Clean the test directory
"""

import sys, os

repo_root = os.path.normpath(os.path.abspath(os.path.join(os.path.dirname(sys.argv[0]), os.pardir)))
os.chdir(repo_root)
sys.path[0] = repo_root

from mmgen.cfg import Config

opts_data = {
	'text': {
		'desc': 'Clean the test directory',
		'usage':'[options]',
		'options': """
-h, --help           Print this help message
--, --longhelp       Print help message for long (global) options
""",
	},
}

cfg = Config(opts_data=opts_data, init_opts={'skip_cfg_file': True})

from test.overlay import get_overlay_tree_dir
overlay_tree_dir = get_overlay_tree_dir(repo_root)
if os.path.exists(overlay_tree_dir):
	from shutil import rmtree
	rmtree(overlay_tree_dir, ignore_errors=True)
	print(f'Removed {os.path.relpath(overlay_tree_dir)!r}')

from test.include.common import clean, set_globals

set_globals(cfg)

extra_dirs = [
	Config.test_datadir,
	os.path.join('test', 'trash'),
	os.path.join('test', 'trash2')
]

from test.cmdtest_d.include.cfg import cfgs
clean(cfgs, extra_dirs=extra_dirs)



#!/usr/bin/env python3

text = {
	'gr': 'Greek text: {}'.format(''.join(map(chr, list(range(913, 939))))),
	'ru': 'Russian text: {}'.format(''.join(map(chr, list(range(1040, 1072))))),
	'zh': 'Chinese text: {}'.format('所以，我們非常需要這樣一種電子支付系統，它基於密碼學原理而不基於信用，'),
	'jp': 'Japanese text: {}'.format('必要なのは、信用ではなく暗号化された証明に基づく電子取引システムであり、')
}

import sys
from mmgen.util import msg, die

if len(sys.argv) != 2 or not sys.argv[1] in text:
	die(2, f'argument must be one of {list(text.keys())}')

msg(text[sys.argv[1]])



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
tool.api: Python interface to selected commands from the 'mmgen-tool' utility
"""

from .common import tool_cmd_base
from .util import tool_cmd as util_cmds
from .coin import tool_cmd as coin_cmds
from .mnemonic import tool_cmd as mnemonic_cmds

class tool_api(
		util_cmds,
		coin_cmds,
		mnemonic_cmds,
		tool_cmd_base):
	"""
	API providing access to a subset of methods from the mmgen.tool module

	Example:
		from mmgen.tool.api import tool_api
		from mmgen.cfg import Config

		# Initialize a tool API instance:
		tool = tool_api(Config())

		# Set the coin and network:
		tool.init_coin('btc', 'mainnet')

		# Print available address types:
		tool.print_addrtypes()

		# Set the address type:
		tool.addrtype = 'segwit'

		# Disable user entropy gathering (optional, reduces security):
		tool.usr_randchars = 0

		# Generate a random BTC segwit keypair:
		wif, addr = tool.randpair()

		# Set coin, network and address type:
		tool.init_coin('ltc', 'testnet')
		tool.addrtype = 'bech32'

		# Generate a random LTC testnet Bech32 keypair:
		wif, addr = tool.randpair()

		print('wif: ', wif)
		print('addr:', addr)
	"""

	need_proto = True
	need_addrtype = True

	def __init__(self, cfg):
		"""
		Initializer - takes no arguments
		"""
		type(cfg)._reset_ok += ('usr_randchars',)
		super().__init__(cfg=cfg)

	def init_coin(self, coinsym, network):
		"""
		Initialize a coin/network pair
		Valid choices for coins: one of the symbols returned by the 'coins' attribute
		Valid choices for network: 'mainnet', 'testnet', 'regtest'
		"""
		from ..protocol import init_proto, warn_trustlevel
		warn_trustlevel(self.cfg)
		self.proto = init_proto(self.cfg, coinsym, network=network, need_amt=True)
		return self.proto

	@property
	def coins(self):
		"""The available coins"""
		from ..protocol import CoinProtocol
		from ..altcoin.params import CoinInfo
		return sorted(set(
			[c.upper() for c in CoinProtocol.coins]
			+ [c.symbol for c in CoinInfo.get_supported_coins(self.proto.network)]
		))

	@property
	def coin(self):
		"""The currently configured coin"""
		return self.proto.coin

	@property
	def network(self):
		"""The currently configured network"""
		return self.proto.network

	@property
	def addrtypes(self):
		"""
		The available address types for current coin/network pair.  The
		first-listed is the default
		"""
		from ..addr import MMGenAddrType
		return [MMGenAddrType(proto=self.proto, id_str=id_str).name for id_str in self.proto.mmtypes]

	def print_addrtypes(self):
		"""
		Print the available address types for current coin/network pair along with
		a description.  The first-listed is the default
		"""
		from ..addr import MMGenAddrType
		for t in [MMGenAddrType(proto=self.proto, id_str=id_str) for id_str in self.proto.mmtypes]:
			print(f'{t.name:<12} - {t.desc}')

	@property
	def addrtype(self):
		"""The currently configured address type (is assignable)"""
		return self.mmtype

	@addrtype.setter
	def addrtype(self, val):
		from ..addr import MMGenAddrType
		self.mmtype = MMGenAddrType(self.proto, val)

	@property
	def usr_randchars(self):
		"""
		The number of keystrokes of entropy to be gathered from the user.
		Setting to zero disables user entropy gathering.
		"""
		return self.cfg.usr_randchars

	@usr_randchars.setter
	def usr_randchars(self, val):
		self.cfg.usr_randchars = val



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
http: HTTP client base class
"""

import requests

class HTTPClient:

	network_proto = 'https'
	host = None
	timeout = 60
	http_hdrs = {
		'User-Agent': 'curl/8.7.1',
		'Proxy-Connection': 'Keep-Alive'}
	extra_http_hdrs = {}
	verify = True
	text_mode = True

	def __init__(self, cfg, *, network_proto=None, host=None):
		self.cfg = cfg
		if network_proto:
			self.network_proto = network_proto
		if host:
			self.host = host
		self.session = requests.Session()
		self.session.trust_env = False # ignore *_PROXY environment vars
		self.session.headers = (self.http_hdrs | self.extra_http_hdrs)
		if cfg.proxy == 'env':
			self.session.trust_env = True
		elif cfg.proxy:
			self.session.proxies.update({
				'http':  f'socks5h://{cfg.proxy}',
				'https': f'socks5h://{cfg.proxy}'
			})

	def call(self, name, path, err_fs, timeout, *, data=None):
		url = self.network_proto + '://' + self.host + path
		kwargs = {
			'url': url,
			'timeout': self.cfg.http_timeout or timeout or self.timeout,
			'verify': self.verify}
		if data:
			kwargs['data'] = data
		res = getattr(self.session, name)(**kwargs)
		if res.status_code != 200:
			from .util import die
			die(2, '\n' + err_fs.format(s=res.status_code, u=url, d=data))
		return res.content.decode() if self.text_mode else res.content

	def get(self, *, path, timeout=None):
		return self.call(
			'get',
			path,
			'HTTP GET failed with status code {s}\n  URL: {u}',
			timeout)

	def post(self, *, path, data, timeout=None):
		return self.call(
			'post',
			path,
			'HTTP POST failed with status code {s}\n  URL: {u}\n  DATA: {d}',
			timeout,
			data = data)



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>

"""
test.objtest_d.btc_mainnet: BTC mainnet test vectors for MMGen data objects
"""

from decimal import Decimal

from mmgen.obj import (
	Int,
	MMGenTxID,
	CoinTxID,
	MMGenWalletLabel,
	MMGenTxComment,
	MMGenPWIDString,
	TrackingWalletName)
from mmgen.addrlist import AddrIdx, AddrIdxList, AddrListID
from mmgen.seed import Seed, SeedID
from mmgen.subseed import SubSeedList, SubSeedIdx, SubSeedIdxRange
from mmgen.seedsplit import SeedShareIdx, SeedShareCount, MasterShareIdx, SeedSplitSpecifier
from mmgen.key import PrivKey, WifKey, PubKey
from mmgen.amt import BTCAmt
from mmgen.addr import CoinAddr, MMGenID, MMGenAddrType, MMGenPasswordType
from mmgen.tw.shared import TwMMGenID, TwLabel, TwComment
from mmgen.rpc.util import IPPort
from mmgen.protocol import init_proto

from .common import r16, r32
from ..include.common import (
	cfg,
	utf8_ctrl,
	utf8_text,
	utf8_combining,
	gr_uc_w_ctrl,
	text_jp,
	text_zh,
	ru_uc,
	gr_uc)
proto = init_proto(cfg, 'btc', need_amt=True)
tw_pfx = proto.base_coin.lower() + ':'
zero_addr = '1111111111111111111114oLvT2'

ssm = str(SeedShareCount.max_val)
privkey = PrivKey(proto=proto, s=bytes.fromhex('deadbeef'*8), compressed=True, pubkey_type='std')

tests = {
	'Int': {
		'arg1': 'n',
		'bad': ('1L', 0.0, '0.0', '1.0', 1.0, 's', 1.1, '1.1'),
		'good': (
			('0', 0), ('-1', -1), ('7', 7), -1, 0, 1, 9999999,
			{'n': '0x0', 'base': 16, 'ret': 0},
			{'n': '0x1', 'base': 16, 'ret': 1},
			{'n': '0xf', 'base': 16, 'ret': 15},
			{'n': '0xff', 'base': 16, 'ret': 255},
		)
	},
	'AddrIdx': {
		'arg1': 'n',
		'bad': ('s', 1.1, 10000000, -1, 0),
		'good': (('7', 7), (1, 1), (9999999, 9999999))
	},
	'SeedShareIdx': {
		'arg1': 'n',
		'bad': ('s', 1.1, 1025, -1, 0),
		'good': (('7', 7), (1, 1), (1024, 1024))
	},
	'SeedShareCount': {
		'arg1': 'n',
		'bad': ('s', 2.1, 1025, -1, 0, 1),
		'good': (('7', 7), (2, 2), (1024, 1024))
	},
	'MasterShareIdx': {
		'arg1': 'n',
		'bad': ('s', 1.1, 1025, -1, 0),
		'good': (('7', 7), (1, 1), (1024, 1024))
	},
	'AddrIdxList': {
		'arg1': 'fmt_str',
		'bad': (
			{'fmt_str': 'x'},
			{'fmt_str': '5,9,1-2-3'},
			{'fmt_str': '8,-11'},
			{'fmt_str': '66,3-2'},
			{'fmt_str': '0-3'},
		),
		'good': (
			{'fmt_str': '3,2,2', 'ret': (2, 3)},
			{'fmt_str': '101,1,3,5,2-7,99', 'ret': (1, 2, 3, 4, 5, 6, 7, 99, 101)},
			{'idx_list': AddrIdxList(fmt_str='1-5'), 'ret': (1, 2, 3, 4, 5)},
		)
	},
	'SubSeedIdxRange': {
		'bad': (33, 'x', '-11', '66,3','0', '3-2', '8000000', '100000000', (1, 2, 3)),
		'good': (
			('3', (3, 3)),
			((3, 5), (3, 5)),
			('1-2', (1, 2)),
			(str(SubSeedList.dfl_len), (SubSeedList.dfl_len, SubSeedList.dfl_len)),
			(str(SubSeedIdxRange.max_idx), (SubSeedIdxRange.max_idx, SubSeedIdxRange.max_idx)),
		)
	},
	'BTCAmt': {
		'arg1': 'num',
		'bad': ('-3.2', '0.123456789', 123, '123L', '22000000', 20999999.12345678,
					{'num': '1', 'from_decimal': True},
					{'num': 1, 'from_decimal': True},
				),
		'good': (('20999999.12345678', Decimal('20999999.12345678')),
				# rounding
				{'num': Decimal('1.23456789623456789'), 'from_decimal': True, 'ret': Decimal('1.23456790')},
				{'num': Decimal('1.234'), 'from_decimal': True, 'ret': Decimal('1.234')},
				{'num': Decimal('0.0'), 'from_decimal': True, 'ret': Decimal('0')},
				# emulate network fee estimation:
				#                  BTC/kB         fee_adjust       tx size
				{   'num': Decimal('0.00053249') * Decimal('0.9') * 109 / 1024 , # ≈53 sat/byte
					'from_decimal': True,
					'ret': Decimal('0.00005101')},
				{   'num': Decimal('0.00003249') * Decimal('1.1') * 109 / 1024 , # ≈3 sat/byte
					'from_decimal': True,
					'ret': Decimal('0.00000380')},
				{   'num': Decimal('0.00011249') * Decimal('1.0') * 221 / 1024 , # ≈11 sat/byte
					'from_decimal': True,
					'ret': Decimal('0.00002428')},
				{'num': 1234, 'from_unit': 'satoshi', 'ret': Decimal('0.00001234')},
		)
	},
	'CoinAddr': {
		'arg1': 'addr',
		'good': (
			{'addr': '1MjjELEy6EJwk8fSNfpS8b5teFRo4X5fZr', 'proto': proto},
			{'addr': '32GiSWo9zJQgkCmjAaLRrbPwXhKry2jHhj', 'proto': proto},
		),
		'bad': (
			{'addr': 1,   'proto': proto},
			{'addr': 'x', 'proto': proto},
			{'addr': 'я', 'proto': proto},
		),
	},
	'SeedID': {
		'arg1': 'sid',
		'bad': (
			{'sid': 'я'},
			{'sid': 'F00F00'},
			{'sid': 'xF00F00x'},
			{'sid': 1},
			{'sid': 'F00BAA123'},
			{'sid': 'f00baa12'},
			{'seed': r32},
			{'sid': 'abc'},
		),
		'good': (
			{'sid': 'F00BAA12'},
			{'seed': Seed(cfg, seed_bin=r16),     'ret': SeedID(seed=Seed(cfg, seed_bin=r16))},
			{'sid':  Seed(cfg, seed_bin=r16).sid, 'ret': SeedID(seed=Seed(cfg, seed_bin=r16))}
		)
	},
	'SubSeedIdx': {
		'arg1': 's',
		'bad': (33, 'x', 'я', '1x', 200, '1ss', 'L', 's', '200LS', '30ll', 's100',
			str(SubSeedIdxRange.max_idx+1), '0'),
		'good': (
			('1', '1L'), ('1s', '1S'), '20S', '30L', ('300l', '300L'), ('200', '200L'),
			str(SubSeedIdxRange.max_idx)+'S')
	},
	'MMGenID': {
		'arg1': 'id_str',
		'bad': (
			{'id_str': 'x',             'proto': proto},
			{'id_str': 1,               'proto': proto},
			{'id_str': 'f00f00f',       'proto': proto},
			{'id_str': 'a:b',           'proto': proto},
			{'id_str': 'x:L:3',         'proto': proto},
			{'id_str': 'F00BAA12',      'proto': proto},
			{'id_str': 'F00BAA12:Z:99', 'proto': proto},
		),
		'good': (
			{'id_str': 'F00BAA12:99',   'proto': proto, 'ret': 'F00BAA12:L:99'},
			{'id_str': 'F00BAA12:L:99', 'proto': proto},
			{'id_str': 'F00BAA12:S:99', 'proto': proto},
		),
	},
	'TwMMGenID': {
		'arg1': 'id_str',
		'bad': (
			{'id_str': 'x',             'proto': proto},
			{'id_str': 'я',             'proto': proto},
			{'id_str': 'я:я',           'proto': proto},
			{'id_str': 1,               'proto': proto},
			{'id_str': 'f00f00f',       'proto': proto},
			{'id_str': 'a:b',           'proto': proto},
			{'id_str': 'x:L:3',         'proto': proto},
			{'id_str': 'F00BAA12:0',    'proto': proto},
			{'id_str': 'F00BAA12:Z:99', 'proto': proto},
			{'id_str': tw_pfx,          'proto': proto},
			{'id_str': tw_pfx+'я',      'proto': proto},
			{'id_str': tw_pfx+'x',      'proto': proto},
		),
		'good': (
			{'id_str': tw_pfx+zero_addr,     'proto': proto},
			{'id_str': 'F00BAA12:99',        'proto': proto, 'ret': 'F00BAA12:L:99'},
			{'id_str': 'F00BAA12:L:99',      'proto': proto},
			{'id_str': 'F00BAA12:S:9999999', 'proto': proto},
		),
	},
	'TrackingWalletName': {
		'arg1': 's',
		'bad': ('-abcdefg', 'a' * 50, 'abc-α-ω'),
		'good': ('ab-cd-def_ghi-', '321-abcd')
	},
	'TwLabel': {
		'arg1': 'proto',
		'exc_name': 'BadTwLabel',
		'bad': (
			{'text': '',              'proto': proto},
			{'text': 'x x',           'proto': proto},
			{'text': 'x я',           'proto': proto},
			{'text': 'я:я',           'proto': proto},
			{'text': 1,               'proto': proto},
			{'text': 'f00f00f',       'proto': proto},
			{'text': 'a:b',           'proto': proto},
			{'text': 'x:L:3',         'proto': proto},
			{'text': 'F00BAA12:0 x',  'proto': proto},
			{'text': 'F00BAA12:Z:99', 'proto': proto},
			{'text': tw_pfx+' x',     'proto': proto},
			{'text': tw_pfx+'я x',    'proto': proto},
			{'text': utf8_ctrl[:40],  'proto': proto},
			{'text': 'F00BAA12:S:1 ' + utf8_ctrl[:40], 'proto': proto, 'exc_name': 'BadTwComment'},
			{'text': tw_pfx+'x comment', 'proto': proto},
		),
		'good': (
			{'text': 'F00BAA12:99 a comment',            'proto': proto, 'ret': 'F00BAA12:L:99 a comment'},
			{'text': 'F00BAA12:L:99 a comment',          'proto': proto},
			{'text': 'F00BAA12:L:99 comment (UTF-8) α',  'proto': proto},
			{'text': 'F00BAA12:S:9999999 comment',       'proto': proto},
			{'text': tw_pfx+zero_addr+' comment',        'proto': proto},
		),
	},
	'MMGenTxID': {
		'arg1': 's',
		'bad': (1, [], '\0', '\1', 'я', 'g', 'gg', 'FF', 'f00', 'F00F0012'),
		'good': ('DEADBE', 'F00BAA')
	},
	'CoinTxID': {
		'arg1': 's',
		'bad': (1, [], '\0', '\1', 'я', 'g', 'gg', 'FF', 'f00', 'F00F0012', r16.hex(), r32.hex()+'ee'),
		'good': (r32.hex(),)
	},
	'WifKey': {
		'arg1': 'proto',
		'bad': (
			{'proto': proto, 'wif': 1},
			{'proto': proto, 'wif': []},
			{'proto': proto, 'wif': '\0'},
			{'proto': proto, 'wif': '\1'},
			{'proto': proto, 'wif': 'я'},
			{'proto': proto, 'wif': 'g'},
			{'proto': proto, 'wif': 'gg'},
			{'proto': proto, 'wif': 'FF'},
			{'proto': proto, 'wif': 'f00'},
			{'proto': proto, 'wif': r16.hex()},
			{'proto': proto, 'wif': '2MspvWFjBbkv2wzQGqhxJUYPCk3Y2jMaxLN'},
		),
		'good': (
			{'proto': proto, 'wif': '5KXEpVzjWreTcQoG5hX357s1969MUKNLuSfcszF6yu84kpsNZKb',  'ret_idx': 1},
			{'proto': proto, 'wif': 'KwWr9rDh8KK5TtDa3HLChEvQXNYcUXpwhRFUPc5uSNnMtqNKLFhk', 'ret_idx': 1},
		)
	},
	'PubKey': {
		'arg1': 's',
		'bad': (
			{'s': 1,          'compressed': True},
			{'s': 'F00BAA12', 'compressed': False},
		),
		'good': ( # TODO: add real pubkeys
			{'s': bytes.fromhex('deadbeef'), 'compressed': True},
		)
	},
	'PrivKey': {
		'arg1': 'proto',
		'bad': (
			{'proto': proto, 'wif': 1},
			{'proto': proto, 'wif': '1'},
			{'proto': proto, 'wif': 'cMsqcmDYZP1LdKgqRh9L4ZRU9br28yvdmTPwW2YQwVSN9aQiMAoR'},
			{'proto': proto, 's': r32, 'wif': '5KXEpVzjWreTcQoG5hX357s1969MUKNLuSfcszF6yu84kpsNZKb'},
			{'proto': proto, 'pubkey_type': 'std'},
			{'proto': proto, 's': r32},
			{'proto': proto, 's': r32, 'compressed': 'yes'},
			{'proto': proto, 's': r32, 'compressed': 'yes', 'pubkey_type': 'std'},
			{'proto': proto, 's': r32, 'compressed': True, 'pubkey_type': 'nonstd'},
			{'proto': proto, 's': r32+b'x', 'compressed': True, 'pubkey_type': 'std'}
		),
		'good': (
			{'proto': proto, 'wif': '5KXEpVzjWreTcQoG5hX357s1969MUKNLuSfcszF6yu84kpsNZKb',
			'ret': bytes.fromhex('e0aef965b905a2fedf907151df8e0a6bac832aa697801c51f58bd2ecb4fd381c')},
			{'proto': proto, 'wif': 'KwWr9rDh8KK5TtDa3HLChEvQXNYcUXpwhRFUPc5uSNnMtqNKLFhk',
			'ret': bytes.fromhex('08d0ed83b64b68d56fa064be48e2385060ed205be2b1e63cd56d218038c3a05f')},
			{'proto': proto, 's': r32, 'compressed': False, 'pubkey_type': 'std', 'ret': r32},
			{'proto': proto, 's': r32, 'compressed': True, 'pubkey_type': 'std', 'ret': r32}
		)
	},
	'AddrListID': {
		'arg1': 'sid',
		'bad': (
			{'sid': SeedID(sid='F00BAA12'), 'mmtype': 'Z', 'ret': 'F00BAA12:Z'},
			{'id_str': 'G00BAA12:B', 'proto': proto},
		),
		'good': (
			{'id_str': 'F00BAA12:B', 'proto': proto, 'ret': 'F00BAA12:B'},
			{'sid': SeedID(sid='F00BAA12'), 'mmtype': proto.addr_type(id_str='S'), 'ret': 'F00BAA12:S'},
			{'sid': SeedID(sid='F00BAA12'), 'mmtype': proto.addr_type(id_str='L'), 'ret': 'F00BAA12:L'},
		)
	},
	'MMGenWalletLabel': {
		'arg1': 's',
		'bad': (utf8_text[:49], utf8_combining[:48], utf8_ctrl[:48], gr_uc_w_ctrl),
		'good': (utf8_text[:48],)
	},
	'TwComment': {
		'exc_name': 'BadTwComment',
		'arg1': 's',
		'bad': (    utf8_combining[:40],
					utf8_ctrl[:40],
					text_jp[:41],
					text_zh[:41],
					gr_uc_w_ctrl,
					utf8_text[:81]),
		'good': (  utf8_text[:80],
					(ru_uc + gr_uc + utf8_text)[:80],
					text_jp[:40],
					text_zh[:40])
	},
	'MMGenTxComment': {
		'arg1': 's',
		'bad': (utf8_text[:73], utf8_combining[:72], utf8_ctrl[:72], gr_uc_w_ctrl),
		'good': (utf8_text[:72],)
	},
	'MMGenPWIDString': { # forbidden = list(u' :/\\')
		'arg1': 's',
		'bad': ('foo/', 'foo:', 'foo:\\'),
		'good': ('qwerty@яяя',)
	},
	'MMGenAddrType': {
		'arg1': 'proto',
		'bad': (
			{'proto': proto, 'id_str': 'U',        'ret': 'L'},
			{'proto': proto, 'id_str': 'z',        'ret': 'L'},
			{'proto': proto, 'id_str': 'xx',       'ret': 'C'},
			{'proto': proto, 'id_str': 'dogecoin', 'ret': 'C'},
		),
		'good': (
			{'proto': proto, 'id_str': 'legacy',     'ret': 'L'},
			{'proto': proto, 'id_str': 'L',          'ret': 'L'},
			{'proto': proto, 'id_str': 'compressed', 'ret': 'C'},
			{'proto': proto, 'id_str': 'C',          'ret': 'C'},
			{'proto': proto, 'id_str': 'segwit',     'ret': 'S'},
			{'proto': proto, 'id_str': 'S',          'ret': 'S'},
			{'proto': proto, 'id_str': 'bech32',     'ret': 'B'},
			{'proto': proto, 'id_str': 'B',          'ret': 'B'}
		)
	},
	'MMGenPasswordType': {
		'arg1': 'proto',
		'bad': (
			{'proto': proto, 'id_str': 'U',        'ret': 'L'},
			{'proto': proto, 'id_str': 'z',        'ret': 'L'},
			{'proto': proto, 'id_str': 'я',        'ret': 'C'},
			{'proto': proto, 'id_str': 1,          'ret': 'C'},
			{'proto': proto, 'id_str': 'passw0rd', 'ret': 'C'},
		),
		'good': (
			{'proto': proto, 'id_str': 'password', 'ret': 'P'},
			{'proto': proto, 'id_str': 'P',        'ret': 'P'},
		)
	},
	'SeedSplitSpecifier': {
		'arg1': 's',
		'bad': ('M', 'αβ:2', 1, '0:1', '1:1', '2:1', '3:2', '1:2000', 'abc:0:2'),
		'good': (
			('1:2', '2:2', 'alice:2:2', 'αβ:2:2', '1:'+ssm, ssm+':'+ssm)
		)
	},
	'IPPort': {
		'arg1': 's',
		'bad': (
			'.0.0.0:0',
			'127.0.0.1',
			'127.0.0.1:',
			'20',
			':20',
			'127.0.0.256:0',
			'127.0.0.1:65536',
			'127.0.0.01:60000',
			'0.0.0.0:00',
		),
		'good': (
			(
				'0.0.0.0:0',
				'0.0.0.1:0',
				'0.0.1.0:0',
				'0.0.1.1:0',
				'0.1.0.0:0',
				'1.0.0.0:0',
				'10.0.0.1:0',
				'127.0.0.1:65535',
				'255.0.0.0:400',
				'255.255.255.255:65535',
			)
		)
	},
}



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.eth.tx.status: Ethereum transaction status class
"""

from ....tx import status as TxBase
from ....util import msg, suf, capfirst

class Status(TxBase.Status):

	async def display(self, *, idx=''):

		def do_return(exitval, message):
			if message:
				msg(message)
			return exitval

		tx = self.tx
		coin_txid = '0x' + getattr(tx, f'coin_txid{idx}')
		tx_desc = 'transaction' + (f' {idx}' if idx else '')

		async def is_in_mempool():
			if not 'full_node' in tx.rpc.caps:
				return False
			if tx.rpc.daemon.id in ('parity', 'openethereum'):
				return coin_txid in [x['hash'] for x in await tx.rpc.call('parity_pendingTransactions')]
			elif tx.rpc.daemon.id in ('geth', 'reth', 'erigon'):
				def gen(key):
					for e in res[key].values():
						for v in e.values():
							yield v['hash']
				res = await tx.rpc.call('txpool_content')
				return coin_txid in list(gen('queued')) + list(gen('pending'))

		async def is_in_wallet():
			d = await tx.rpc.call('eth_getTransactionReceipt', coin_txid)
			if d and 'blockNumber' in d and d['blockNumber'] is not None:
				from collections import namedtuple
				receipt_info = namedtuple('receipt_info', ['confs', 'exec_status', 'rx'])
				return receipt_info(
					confs = 1 + int(await tx.rpc.call('eth_blockNumber'), 16) - int(d['blockNumber'], 16),
					exec_status = int(d['status'], 16),
					rx = d)

		if await is_in_mempool():
			return do_return(0, f'{capfirst(tx_desc)} is in mempool')

		if res := await is_in_wallet():
			if tx.txobj['data'] and not tx.is_swap:
				cd = capfirst(tx.contract_desc)
				msg(f'{cd} failed to execute!' if res.exec_status == 0 else
					f'{cd} successfully executed with status {res.exec_status}')
			return do_return(
				int(not res.exec_status),
				f'{capfirst(tx_desc)} has {res.confs} confirmation{suf(res.confs)}')

		return do_return(1, f'{capfirst(tx_desc)} is neither in mempool nor blockchain!')

class TokenStatus(Status):
	pass



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
test.cmdtest_d.ref_altcoin: Altcoin reference file tests for the cmdtest.py test suite
"""

from mmgen.color import set_vt100

from .include.common import pwfile, dfl_wpasswd, ref_dir, dfl_words_file, dfl_addr_idx_list
from ..include.common import joinpath, start_test_daemons, stop_test_daemons, cmp_or_die
from .ref import CmdTestRef
from .base import CmdTestBase

class CmdTestRefAltcoin(CmdTestRef, CmdTestBase):
	'saved and generated altcoin reference files'
	tmpdir_nums = [8]
	networks = ('btc',)
	chk_data = {
		'ref_addrfile_chksum_zec':        '903E 7225 DD86 6E01',
		'ref_addrfile_chksum_zec_z':      '9C7A 72DC 3D4A B3AF',
		'ref_addrfile_chksum_xmr':        '4369 0253 AC2C 0E38',
		'ref_addrfile_chksum_dash':       'FBC1 6B6A 0988 4403',
		'ref_addrfile_chksum_eth':        'E554 076E 7AF6 66A3',
		'ref_addrfile_chksum_etc':        'E97A D796 B495 E8BC',
		'ref_keyaddrfile_chksum_zec':     'F05A 5A5C 0C8E 2617',
		'ref_keyaddrfile_chksum_zec_z':   '6B87 9B2D 0D8D 8D1E',
		'ref_keyaddrfile_chksum_xmr':     'E0D7 9612 3D67 404A',
		'ref_viewkeyaddrfile_chksum_xmr': '40C9 0E61 B743 229C',
		'ref_keyaddrfile_chksum_dash':    'E83D 2C63 FEA2 4142',
		'ref_keyaddrfile_chksum_eth':     'E400 70D9 0AE3 C7C2',
		'ref_keyaddrfile_chksum_etc':     'EF49 967D BD6C FE45',
	}
	cmd_group = (
		('ref_altcoin_tx_chk',          'signing saved reference tx files'),
		('ref_addrfile_gen_eth',        'generate address file (ETH)'),
		('ref_addrfile_gen_etc',        'generate address file (ETC)'),
		('ref_addrfile_gen_dash',       'generate address file (DASH)'),
		('ref_addrfile_gen_zec',        'generate address file (ZEC-T)'),
		('ref_addrfile_gen_zec_z',      'generate address file (ZEC-Z)'),
		('ref_addrfile_gen_xmr',        'generate address file (XMR)'),

		('ref_keyaddrfile_gen_eth',     'generate key-address file (ETH)'),
		('ref_keyaddrfile_gen_etc',     'generate key-address file (ETC)'),
		('ref_keyaddrfile_gen_dash',    'generate key-address file (DASH)'),
		('ref_keyaddrfile_gen_zec',     'generate key-address file (ZEC-T)'),
		('ref_keyaddrfile_gen_zec_z',   'generate key-address file (ZEC-Z)'),
		('ref_keyaddrfile_gen_xmr',     'generate key-address file (XMR)'),
		('ref_viewkeyaddrfile_gen_xmr', 'generate viewkey-address file (XMR)'),

		('ref_addrfile_chk_eth',        'reference address file (ETH)'),
		('ref_addrfile_chk_etc',        'reference address file (ETC)'),
		('ref_addrfile_chk_dash',       'reference address file (DASH)'),
		('ref_addrfile_chk_zec',        'reference address file (ZEC-T)'),
		('ref_addrfile_chk_zec_z',      'reference address file (ZEC-Z)'),
		('ref_addrfile_chk_xmr',        'reference address file (XMR)'),

		('ref_keyaddrfile_chk_eth',     'reference key-address file (ETH)'),
		('ref_keyaddrfile_chk_etc',     'reference key-address file (ETC)'),
		('ref_keyaddrfile_chk_dash',    'reference key-address file (DASH)'),
		('ref_keyaddrfile_chk_zec',     'reference key-address file (ZEC-T)'),
		('ref_keyaddrfile_chk_zec_z',   'reference key-address file (ZEC-Z)'),
		('ref_keyaddrfile_chk_xmr',     'reference key-address file (XMR)'),
		('ref_viewkeyaddrfile_chk_xmr', 'reference viewkey-address file (XMR)'),
	)

	def ref_altcoin_tx_chk(self):
		"""
		Check saved transaction files for *all* configured altcoins
		Though this basically duplicates the autosign test, here we do everything
		via the command line, so it's worth doing
		"""
		self.write_to_tmpfile(pwfile, dfl_wpasswd)
		passfile = joinpath(self.tmpdir, pwfile)
		from mmgen.tx.file import MMGenTxFile
		src = CmdTestRef.sources['ref_tx_file']
		for coin, files in src.items():
			if coin == 'mm1':
				coin = 'eth'
				token_desc = ':MM1'
			else:
				token_desc = ''
			for fn in files:
				if not fn: # no etc testnet TX file
					continue
				txfile = joinpath(
					ref_dir,
					self._get_ref_subdir_by_coin(coin),
					fn)
				proto = MMGenTxFile.get_proto(self.cfg, txfile, quiet_open=True)
				if proto.sign_mode == 'daemon':
					start_test_daemons(proto.network_id)
					set_vt100()
				t = self.spawn(
					'mmgen-txsign',
					['--outdir=test/trash', '--yes', f'--passwd-file={passfile}', dfl_words_file, txfile],
					no_passthru_opts = ['coin'],
					extra_desc = f'{proto.coin}{token_desc} {proto.network}')
				t.read()
				t.ok()
				if proto.sign_mode == 'daemon' and not proto.network_id == 'btc':
					stop_test_daemons(proto.network_id, remove_datadir=True)
		return 'ok'

	def ref_altcoin_addrgen(
			self,
			coin,
			mmtype,
			gen_what      = 'addr',
			coin_suf      = '',
			add_args      = [],
			addr_idx_list = None):
		wf = dfl_words_file
		t = self.spawn(
				'mmgen-keygen' if 'key' in gen_what else 'mmgen-addrgen',
				['-Sq', '--coin='+coin] +
				(['--type='+mmtype] if mmtype else []) +
				add_args +
				[wf, addr_idx_list or dfl_addr_idx_list])
		if 'key' in gen_what:
			t.expect('Encrypt key list? (y/N): ', 'N')
		chk = t.expect_getend(r'.* data checksum for \S*: ', regex=True)
		chk_ref = self.chk_data[
			'ref_{}addrfile_chksum_{}{}'.format(
				(gen_what if 'key' in gen_what else ''),
				coin.lower(),
				coin_suf)
		]
		cmp_or_die(chk, chk_ref, desc=f'{gen_what}list data checksum')
		return t

	def ref_addrfile_gen_eth(self):
		return self.ref_altcoin_addrgen(coin='ETH', mmtype='ethereum')

	def ref_addrfile_gen_etc(self):
		return self.ref_altcoin_addrgen(coin='ETC', mmtype='ethereum')

	def ref_addrfile_gen_dash(self):
		return self.ref_altcoin_addrgen(coin='DASH', mmtype='compressed')

	def ref_addrfile_gen_zec(self):
		return self.ref_altcoin_addrgen(coin='ZEC', mmtype='compressed')

	def ref_addrfile_gen_zec_z(self):
		return self.ref_altcoin_addrgen(coin='ZEC', mmtype='zcash_z', coin_suf='_z')

	def ref_addrfile_gen_xmr(self):
		return self.ref_altcoin_addrgen(coin='XMR', mmtype='monero')

	def ref_addrfile_gen_xmr_slow(self):
		return self.ref_altcoin_addrgen(coin='XMR', mmtype='monero', add_args=['--keygen-backend=2'])

	def ref_keyaddrfile_gen_eth(self):
		return self.ref_altcoin_addrgen(coin='ETH', mmtype='ethereum', gen_what='key')

	def ref_keyaddrfile_gen_etc(self):
		return self.ref_altcoin_addrgen(coin='ETC', mmtype='ethereum', gen_what='key')

	def ref_keyaddrfile_gen_dash(self):
		return self.ref_altcoin_addrgen(coin='DASH', mmtype='compressed', gen_what='key')

	def ref_keyaddrfile_gen_zec(self):
		return self.ref_altcoin_addrgen(coin='ZEC', mmtype='compressed', gen_what='key')

	def ref_keyaddrfile_gen_zec_z(self):
		return self.ref_altcoin_addrgen(coin='ZEC', mmtype='zcash_z', coin_suf='_z', gen_what='key')

	def ref_keyaddrfile_gen_xmr(self):
		return self.ref_altcoin_addrgen(coin='XMR', mmtype='monero', gen_what='key')

	def ref_viewkeyaddrfile_gen_xmr(self):
		return self.ref_altcoin_addrgen(
			coin          = 'XMR',
			mmtype        = 'monero',
			gen_what      = 'viewkey',
			add_args      = ['--viewkeys'],
			addr_idx_list = '1-3')

	def ref_addrfile_chk_eth(self):
		return self.ref_addrfile_chk(ftype='addr', coin='ETH', subdir='ethereum', pfx='-ETH',
				pat='ETH Mainnet.*Ethereum')

	def ref_addrfile_chk_etc(self):
		return self.ref_addrfile_chk(ftype='addr', coin='ETC', subdir='ethereum_classic', pfx='-ETC',
				pat='ETH Mainnet.*Ethereum')

	def ref_addrfile_chk_dash(self):
		return self.ref_addrfile_chk(ftype='addr', coin='DASH', subdir='dash', pfx='-DASH-C',
				pat='DASH Mainnet.*Compressed')

	def ref_addrfile_chk_zec(self):
		return self.ref_addrfile_chk(ftype='addr', coin='ZEC', subdir='zcash', pfx='-ZEC-C',
				pat='ZEC Mainnet.*Compressed')

	def ref_addrfile_chk_zec_z(self):
		return self.ref_addrfile_chk(ftype='addr', coin='ZEC', subdir='zcash', pfx='-ZEC-Z', mmtype='z',
				pat='ZEC Mainnet.*Zcash_z')

	def ref_addrfile_chk_xmr(self):
		return self.ref_addrfile_chk(ftype='addr', coin='XMR', subdir='monero', pfx='-XMR-M',
				pat='XMR Mainnet.*Monero')

	def ref_keyaddrfile_chk_eth(self):
		return self.ref_addrfile_chk(ftype='keyaddr', coin='ETH', subdir='ethereum', pfx='-ETH',
				pat='ETH Mainnet.*Ethereum')

	def ref_keyaddrfile_chk_etc(self):
		return self.ref_addrfile_chk(ftype='keyaddr', coin='ETC', subdir='ethereum_classic', pfx='-ETC',
				pat='ETH Mainnet.*Ethereum')

	def ref_keyaddrfile_chk_dash(self):
		return self.ref_addrfile_chk(ftype='keyaddr', coin='DASH', subdir='dash', pfx='-DASH-C',
				pat='DASH Mainnet.*Compressed')

	def ref_keyaddrfile_chk_zec(self):
		return self.ref_addrfile_chk(ftype='keyaddr', coin='ZEC', subdir='zcash', pfx='-ZEC-C',
				pat='ZEC Mainnet.*Compressed')

	def ref_keyaddrfile_chk_zec_z(self):
		return self.ref_addrfile_chk(ftype='keyaddr', coin='ZEC', subdir='zcash', pfx='-ZEC-Z', mmtype='z',
				pat='ZEC Mainnet.*Zcash_z')

	def ref_keyaddrfile_chk_xmr(self):
		return self.ref_addrfile_chk(ftype='keyaddr', coin='XMR', subdir='monero', pfx='-XMR-M',
				pat='XMR Mainnet.*Monero')

	def ref_viewkeyaddrfile_chk_xmr(self):
		return self.ref_addrfile_chk(ftype='viewkeyaddr', coin='XMR', subdir='monero', pfx='-XMR-M',
				pat='XMR Mainnet.*Monero')



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
mmgen-txsend: Broadcast a transaction signed by 'mmgen-txsign' to the network
"""

import sys

from .cfg import gc, Config
from .util import async_run, die

opts_data = {
	'sets': [
		('yes', True, 'quiet', True),
		('abort', True, 'autosign', True),
	],
	'text': {
		'desc':    f'Send a signed {gc.proj_name} cryptocoin transaction',
		'usage':   '[opts] [signed transaction file]',
		'options': """
-h, --help       Print this help message
--, --longhelp   Print help message for long (global) options
-a, --autosign   Send an autosigned transaction created by ‘mmgen-txcreate
                 --autosign’.  The removable device is mounted and unmounted
                 automatically. The transaction file argument must be omitted
                 when using this option
-A, --abort      Abort an unsent transaction created by ‘mmgen-txcreate
                 --autosign’ and delete it from the removable device.  The
                 transaction may be signed or unsigned.
-d, --outdir=d   Specify an alternate directory 'd' for output
-H, --dump-hex=F Instead of sending to the network, dump the transaction hex
                 to file ‘F’.  Use filename ‘-’ to dump to standard output.
-m, --mark-sent  Mark the transaction as sent by adding it to the removable
                 device.  Used in combination with --autosign when a trans-
                 action has been successfully sent out-of-band.
-n, --tx-proxy=P Send transaction via public TX proxy ‘P’ (supported proxies:
                 {tx_proxies}).  This is done via a publicly accessible web
                 page, so no API key or registration is required.
-q, --quiet      Suppress warnings; overwrite files without prompting
-r, --receipt    Print the receipt of the sent transaction (Ethereum only)
-s, --status     Get status of a sent transaction (or current transaction,
                 whether sent or unsent, when used with --autosign)
-t, --test       Test whether the transaction can be sent without sending it
-T, --txhex-idx=N Send only part ‘N’ of a multi-part transaction.  Indexing
                 begins with one.
-v, --verbose    Be more verbose
-w, --wait       Wait for transaction confirmation (Ethereum only)
-x, --proxy=P    Connect to TX proxy via SOCKS5h proxy ‘P’ (host:port).
                 Use special value ‘env’ to honor *_PROXY environment vars
                 instead.
-y, --yes        Answer 'yes' to prompts, suppress non-essential output
"""
	},
	'code': {
		'options': lambda cfg, proto, help_notes, s: s.format(
			tx_proxies = help_notes('tx_proxies'))
	}
}

cfg = Config(opts_data=opts_data)

if cfg.autosign and cfg.outdir:
	die(1, '--outdir cannot be used in combination with --autosign')

if cfg.mark_sent and not cfg.autosign:
	die(1, '--mark-sent is used only in combination with --autosign')

if cfg.test and cfg.dump_hex:
	die(1, '--test cannot be used in combination with --dump-hex')

if cfg.dump_hex and cfg.dump_hex != '-':
	from .fileutil import check_outfile_dir
	check_outfile_dir(cfg.dump_hex)

asi = None

if len(cfg._args) == 1:
	infile = cfg._args[0]
	from .fileutil import check_infile
	check_infile(infile)
elif not cfg._args and cfg.autosign:
	from .tx.util import mount_removable_device
	from .autosign import Signable
	asi = mount_removable_device(cfg)
	si = Signable.automount_transaction(asi)
	if cfg.abort:
		si.shred_abortable() # prompts user, then raises exception or exits
	elif cfg.status or cfg.receipt:
		if si.unsent:
			die(1, 'Transaction is unsent')
		if si.unsigned:
			die(1, 'Transaction is unsigned')
	else:
		infile = si.get_unsent()
		cfg._util.qmsg(f'Got signed transaction file ‘{infile}’')
else:
	cfg._usage()

if not cfg.status:
	from .ui import do_license_msg
	do_license_msg(cfg)

from .tx import OnlineSignedTX

async def main():

	global cfg

	if (cfg.status or cfg.receipt) and cfg.autosign:
		tx = await si.get_last_created()
	else:
		tx = await OnlineSignedTX(
			cfg        = cfg,
			filename   = infile,
			automount  = cfg.autosign,
			quiet_open = True)

	cfg = Config({'_clone': cfg, 'proto': tx.proto, 'coin': tx.proto.coin})

	if cfg.tx_proxy:
		from .tx.tx_proxy import check_client
		check_client(cfg)

	from .rpc import rpc_init
	tx.rpc = await rpc_init(cfg)

	cfg._util.vmsg(f'Getting {tx.desc} ‘{tx.infile}’')

	if cfg.mark_sent:
		await tx.post_send(asi)
		sys.exit(0)

	if not (cfg.status or cfg.receipt):
		if tx.is_swap and not tx.check_swap_expiry():
			die(1, 'Swap quote has expired. Please re-create the transaction')

		if not cfg.yes:
			tx.info.view_with_prompt('View transaction details?')
			if tx.add_comment(): # edits an existing comment, returns true if changed
				if not cfg.autosign:
					tx.file.write(ask_write_default_yes=True)

	await tx.send(cfg, asi)

async_run(main())



#!/usr/bin/env python3

"""
scripts/gendiff.py:

Clean up control characters and trailing whitespace in the listed source files
and create a unified diff between them.

If more or less than two files are listed on the command line, the cleanup is
performed on all files, but no diff is created.

The source files are assumed to be terminal output captured by the `script`
command.

The cleaned source files are saved with the .clean extension.
"""

import sys, re
from difflib import unified_diff

fns = sys.argv[1:3]
diff_opts = sys.argv[4:] if sys.argv[3:4] == ['--'] else None

translate = {
	'\r': None,
	'\b': '[BS]',
#	chr(4): '', # Ctrl-D, EOT
}

def cleanup_file(fn):

	# must use binary mode to prevent conversion of DOS CR into newline
	with open(fn, 'rb') as fp:
		data = fp.read().decode()

	def gen_text():
		for line in data.split('\n'): # do not use splitlines()
			line = line.translate({ord(a):b for a, b in translate.items()})
			line = re.sub(r'\s+$', '', line)  # trailing whitespace
			yield line

	ret = list(gen_text())

	sys.stderr.write(f'Saving cleaned file to {fn}.clean\n')

	with open(f'{fn}.clean', 'w') as fp:
		fp.write('\n'.join(ret))

	return ret

if len(fns) != 2:
	sys.stderr.write(f'{len(fns)} input files.  Not generating diff.\n')

cleaned_texts = [cleanup_file(fn) for fn in fns]

if len(fns) == 2:
	# chunk headers have trailing newlines, hence the rstrip()
	sys.stderr.write('Generating diff\n')
	if diff_opts:
		from subprocess import run
		run(['diff', '-u'] + [f'{fn}.clean' for fn in fns])
	else:
		print(
			f'diff a/{fns[0]} b/{fns[1]}\n' +
			'\n'.join(a.rstrip() for a in unified_diff(
				*cleaned_texts,
				fromfile = f'a/{fns[0]}',
				tofile   = f'b/{fns[1]}'))
		)



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
cfg: Configuration classes for the MMGen suite
"""

import sys, os
from collections import namedtuple
from .base_obj import Lockable

def die(*args, **kwargs):
	from .util import die
	die(*args, **kwargs)

def die2(exit_val, s):
	sys.stderr.write(s+'\n')
	sys.exit(exit_val)

class GlobalConstants(Lockable):
	"""
	These values are non-runtime-configurable.  They’re constant for a given machine,
	user, executable and MMGen Wallet version
	"""
	_autolock = True

	proj_name          = 'MMGen'
	proj_id            = 'mmgen'
	proj_url           = 'https://github.com/mmgen/mmgen-wallet'
	author             = 'The MMGen Project'
	email              = '<mmgen@tuta.io>'
	Cdates             = '2013-2025'
	dfl_hash_preset    = '3'
	passwd_max_tries   = 5
	min_screen_width   = 80
	min_time_precision = 18

	# core_coins must match CoinProtocol.coins
	core_coins = ('btc', 'bch', 'ltc', 'eth', 'etc', 'zec', 'xmr', 'rune')
	rpc_coins = ('btc', 'bch', 'ltc', 'eth', 'etc', 'xmr', 'rune')
	local_rpc_coins = ('btc', 'bch', 'ltc', 'eth', 'etc', 'xmr')
	remote_rpc_coins = ('rune',)
	btc_fork_rpc_coins = ('btc', 'bch', 'ltc')
	eth_fork_coins = ('eth', 'etc')

	# ‘use_coin_opt’ must be False if ‘coin_codes’ is set
	_cc = namedtuple('cmd_cap', ['proto', 'rpc', 'use_coin_opt', 'coin_codes', 'caps', 'platforms'])
	cmd_caps_data = {
		'addrgen':      _cc(True,  False, True,  None,     [],      'lmw'),
		'addrimport':   _cc(True,  True,  True,  None,     ['tw'],  'lmw'),
		'autosign':     _cc(True,  True,  False, '-bRrXx', ['rpc'], 'lm'),
		'cli':          _cc(True,  True,  True,  None,     ['tw'],  'lmw'),
		'keygen':       _cc(True,  False, True,  None,     [],      'lmw'),
		'msg':          _cc(True,  True,  True,  None,     ['msg'], 'lmw'),
		'passchg':      _cc(False, False, False, None,     [],      'lmw'),
		'passgen':      _cc(False, False, False, None,     [],      'lmw'),
		'regtest':      _cc(True,  True,  True,  None,     ['tw'],  'lmw'),
		'seedjoin':     _cc(False, False, False, None,     [],      'lmw'),
		'seedsplit':    _cc(False, False, False, None,     [],      'lmw'),
		'subwalletgen': _cc(False, False, False, None,     [],      'lmw'),
		'swaptxcreate': _cc(True,  True,  False, '-bRrx',  ['tw'],  'lmw'),
		'swaptxdo':     _cc(True,  True,  False, '-bRrx',  ['tw'],  'lmw'),
		'tool':         _cc(True,  True,  True,  None,     [],      'lmw'),
		'txbump':       _cc(True,  True,  True,  None,     ['tw'],  'lmw'),
		'txcreate':     _cc(True,  True,  True,  None,     ['tw'],  'lmw'),
		'txdo':         _cc(True,  True,  True,  None,     ['tw'],  'lmw'),
		'txsend':       _cc(True,  True,  False, '-bRrXx', ['tw'],  'lmw'),
		'txsign':       _cc(True,  True,  False, '-bRrXx', ['tw'],  'lmw'),
		'walletchk':    _cc(False, False, False, None,     [],      'lmw'),
		'walletconv':   _cc(False, False, False, None,     [],      'lmw'),
		'walletgen':    _cc(False, False, False, None,     [],      'lmw'),
		'xmrwallet':    _cc(True,  True,  False, '-rx',    ['rpc'], 'lmw'),
	}
	altcoin_cmds = ('swaptxcreate', 'swaptxdo', 'xmrwallet')

	prog_name = os.path.basename(sys.argv[0])
	prog_id = prog_name.removeprefix(f'{proj_id}-')
	cmd_caps = cmd_caps_data.get(prog_id)

	if sys.platform not in ('linux', 'win32', 'darwin'):
		die2(1, f'{sys.platform!r}: platform not supported by {proj_name}')

	if os.getenv('HOME'):   # Linux, MSYS2, or macOS
		home_dir = os.getenv('HOME')
	elif sys.platform == 'win32': # Windows without MSYS2 - not supported
		die2(1, f'$HOME not set!  {proj_name} for Windows must be run in MSYS2 environment')
	else:
		die2(2, '$HOME is not set!  Unable to determine home directory')

	def get_mmgen_data_file(self, *, filename, package='mmgen'):
		"""
		this is an expensive import, so do only when required
		"""
		# Resource will be unpacked and then cleaned up if necessary, see:
		#    https://docs.python.org/3/library/importlib.html:
		#        Note: This module provides functionality similar to pkg_resources Basic
		#        Resource Access without the performance overhead of that package.
		#    https://importlib-resources.readthedocs.io/en/latest/migration.html
		#    https://setuptools.readthedocs.io/en/latest/pkg_resources.html

		try:
			from importlib.resources import files # Python 3.9 and above
		except ImportError:
			from importlib_resources import files
		return files(package).joinpath('data', filename).read_text()

	@property
	def version(self):
		return self.get_mmgen_data_file(
				filename = 'version',
				package  = 'mmgen_node_tools' if self.prog_name.startswith('mmnode-') else 'mmgen'
			).strip()

	@property
	def release_date(self):
		return self.get_mmgen_data_file(filename='release_date').strip()

gc = GlobalConstants()

class GlobalVars:
	"""
	These are used only by the test suite to redirect msg() and friends to /dev/null
	"""
	stdout = sys.stdout
	stderr = sys.stderr

gv = GlobalVars()

class Config(Lockable):
	"""
	These values are configurable - RHS values are defaults
	Globals are overridden with the following precedence:
	  1 - command line
	  2 - environmental vars
	  3 - config file
	"""
	_autolock = False
	_set_ok = ('usr_randchars', '_proto')
	_reset_ok = ('accept_defaults',)
	_delete_ok = ('_opts',)
	_use_class_attr = True
	_default_to_none = True

	# general
	coin        = 'BTC'
	token       = ''
	outdir      = ''
	passwd_file = ''
	network     = 'mainnet'
	testnet     = False
	regtest     = False

	# verbosity / prompting behavior
	quiet           = False
	verbose         = False
	yes             = False
	accept_defaults = False
	no_license      = False

	# limits
	http_timeout       = 0
	daemon_state_timeout = 60
	usr_randchars      = 30
	fee_adjust         = 1.0
	fee_estimate_confs = 3
	minconf            = 1
	max_tx_file_size   = 100000
	max_input_size     = 1024 * 1024
	min_urandchars     = 10
	max_urandchars     = 80
	macos_autosign_ramdisk_size = 10 # see MacOSRamDisk

	# debug
	debug                = False
	debug_daemon         = False
	debug_evm            = False
	debug_opts           = False
	debug_rpc            = False
	debug_addrlist       = False
	debug_subseed        = False
	debug_tw             = False
	devtools             = False
	traceback            = False

	# rpc:
	rpc_host              = ''
	rpc_port              = 0
	rpc_user              = ''
	rpc_password          = ''
	aiohttp_rpc_queue_len = 16
	cached_balances       = False

	# daemons
	daemon_data_dir       = '' # set by user
	daemon_id             = ''
	blacklisted_daemons   = ''
	ignore_daemon_version = False

	# display:
	test_suite_enable_color = False # placeholder
	force_256_color = False
	scroll          = False
	pager           = False
	columns         = 0
	color = bool(
		(sys.stdout.isatty() and not os.getenv('MMGEN_TEST_SUITE_PEXPECT')) or
		os.getenv('MMGEN_TEST_SUITE_ENABLE_COLOR')
	)

	# miscellaneous features:
	use_internal_keccak_module     = False
	force_standalone_scrypt_module = False
	enable_erigon                  = False
	autochg_ignore_labels          = False
	autosign                       = False

	# regtest:
	bob          = False
	alice        = False
	carol        = False
	regtest_user = ''

	# altcoin:
	cashaddr = True

	# Monero:
	monero_wallet_rpc_user     = 'monero'
	monero_wallet_rpc_password = ''
	priority                   = 0

	# test suite:
	bogus_send               = False
	bogus_unspent_data       = ''
	debug_utf8               = False
	exec_wrapper             = False
	ignore_test_py_exception = False
	test_suite               = False
	test_suite_autosign_led_simulate = False
	test_suite_autosign_threaded = False
	test_suite_devnet_block_period = 0
	test_suite_xmr_autosign  = False
	test_suite_cfgtest       = False
	test_suite_deterministic = False
	test_suite_pexpect       = False
	test_suite_pexpect_timeout = 0
	test_suite_popen_spawn   = False
	test_suite_root_pfx      = ''
	hold_protect_disable     = False
	no_daemon_autostart      = False
	names                    = False
	no_timings               = False
	exit_after               = ''
	resuming                 = False
	skipping_deps            = False
	test_datadir             = os.path.join('test', 'data_dir' + ('', '-α')[bool(os.getenv('MMGEN_DEBUG_UTF8'))])

	mnemonic_entry_modes = {}

	# external use:
	_opts  = None
	_proto = None

	# internal use:
	_use_cfg_file           = False
	_use_env                = False

	_forbidden_opts = (
		'data_dir_root',
	)

	_incompatible_opts = (
		('help', 'longhelp'),
		('bob', 'alice', 'carol'),
		('label', 'keep_label'),
		('tx_id', 'info'),
		('tx_id', 'terse_info'),
		('autosign', 'outdir'),
	)

	# proto-specific only: eth_mainnet_chain_names eth_testnet_chain_names
	# coin-specific only:  bch_cashaddr (alias of cashaddr)
	_cfg_file_opts = (
		'autochg_ignore_labels',
		'autosign',
		'color',
		'daemon_data_dir',
		'daemon_id', # also coin-specific
		'debug',
		'fee_adjust',
		'force_256_color',
		'hash_preset',
		'http_timeout',
		'ignore_daemon_version', # also coin-specific
		'macos_autosign_ramdisk_size',
		'max_input_size',
		'max_tx_file_size',
		'mnemonic_entry_modes',
		'monero_wallet_rpc_password',
		'monero_wallet_rpc_user',
		'no_license',
		'quiet',
		'regtest',
		'rpc_host',     # also coin-specific
		'rpc_password', # also coin-specific
		'rpc_port',     # also coin-specific
		'rpc_user',     # also coin-specific
		'scroll',
		'subseeds',
		'testnet',
		'tw_name',      # also coin-specific
		'usr_randchars')

	# Supported environmental vars
	# The corresponding attributes (lowercase, without 'mmgen_') must exist in the class.
	# The 'MMGEN_DISABLE_' prefix sets the corresponding attribute to False.
	_env_opts = (
		'MMGEN_DEBUG_ALL', # special: there is no `debug_all` attribute

		'MMGEN_COLUMNS',
		'MMGEN_TEST_SUITE',
		'MMGEN_TEST_SUITE_AUTOSIGN_LED_SIMULATE',
		'MMGEN_TEST_SUITE_AUTOSIGN_THREADED',
		'MMGEN_TEST_SUITE_DEVNET_BLOCK_PERIOD',
		'MMGEN_TEST_SUITE_XMR_AUTOSIGN',
		'MMGEN_TEST_SUITE_CFGTEST',
		'MMGEN_TEST_SUITE_DETERMINISTIC',
		'MMGEN_TEST_SUITE_ENABLE_COLOR',
		'MMGEN_TEST_SUITE_PEXPECT',
		'MMGEN_TEST_SUITE_PEXPECT_TIMEOUT',
		'MMGEN_TEST_SUITE_POPEN_SPAWN',
		'MMGEN_TEST_SUITE_ROOT_PFX',
		'MMGEN_TRACEBACK',
		'MMGEN_BLACKLIST_DAEMONS',
		'MMGEN_BOGUS_SEND',
		'MMGEN_BOGUS_UNSPENT_DATA',
		'MMGEN_DAEMON_STATE_TIMEOUT',
		'MMGEN_DEBUG',
		'MMGEN_DEBUG_DAEMON',
		'MMGEN_DEBUG_EVM',
		'MMGEN_DEBUG_OPTS',
		'MMGEN_DEBUG_RPC',
		'MMGEN_DEBUG_ADDRLIST',
		'MMGEN_DEBUG_TW',
		'MMGEN_DEBUG_UTF8',
		'MMGEN_DEBUG_SUBSEED',
		'MMGEN_DEVTOOLS',
		'MMGEN_FORCE_256_COLOR',
		'MMGEN_HOLD_PROTECT_DISABLE',
		'MMGEN_HTTP_TIMEOUT',
		'MMGEN_QUIET',
		'MMGEN_NO_LICENSE',
		'MMGEN_RPC_HOST',
		'MMGEN_RPC_FAIL_ON_COMMAND',
		'MMGEN_TESTNET',
		'MMGEN_REGTEST',
		'MMGEN_EXEC_WRAPPER',
		'MMGEN_IGNORE_TEST_PY_EXCEPTION',
		'MMGEN_RPC_BACKEND',
		'MMGEN_IGNORE_DAEMON_VERSION',
		'MMGEN_USE_STANDALONE_SCRYPT_MODULE',
		'MMGEN_ENABLE_ERIGON',
		'MMGEN_DISABLE_COLOR',
	)

	_infile_opts = (
		'keys_from_file',
		'mmgen_keys_from_file',
		'passwd_file',
		'keysforaddrs',
		'comment_file',
		'contract_data',
	)

	# Auto-typechecked and auto-set opts - first value in list is the default
	_ov = namedtuple('autoset_opt_info', ['type', 'choices'])
	_autoset_opts = {
		'fee_estimate_mode': _ov('nocase_pfx', ['conservative', 'economical']),
		'rpc_backend':       _ov('nocase_pfx', ['auto', 'httplib', 'curl', 'aiohttp', 'requests']),
		'swap_proto':        _ov('nocase_pfx', ['thorchain']),
		'tx_proxy':          _ov('nocase_pfx', ['etherscan']) # , 'blockchair'
	}
	_dfl_none_autoset_opts = ('tx_proxy',)

	_auto_typeset_opts = {
		'seed_len': int,
		'subseeds': int,
		'vsize_adj': float,
	}

	# test suite:
	err_disp_timeout   = 0.7
	short_disp_timeout = 0.3
	stdin_tty          = sys.stdin.isatty()

	if os.getenv('MMGEN_TEST_SUITE'):
		min_urandchars = 3
		err_disp_timeout = 0.1
		short_disp_timeout = 0.1
		if os.getenv('MMGEN_TEST_SUITE_POPEN_SPAWN'):
			stdin_tty = True
		if gc.prog_name == 'modtest.py':
			_set_ok += ('debug_subseed',)
			_reset_ok += ('force_standalone_scrypt_module',)

	if os.getenv('MMGEN_DEBUG_ALL'):
		for name in _env_opts:
			if name[:11] == 'MMGEN_DEBUG':
				os.environ[name] = '1'

	@property
	def data_dir_root(self):
		"""
		location of mmgen.cfg
		"""
		if not hasattr(self, '_data_dir_root'):
			if self._data_dir_root_override:
				self._data_dir_root = os.path.normpath(os.path.abspath(self._data_dir_root_override))
			elif self.test_suite:
				self._data_dir_root = self.test_datadir
			else:
				self._data_dir_root = os.path.join(gc.home_dir, '.'+gc.proj_name.lower())
		return self._data_dir_root

	@property
	def data_dir(self):
		"""
		location of wallet and other data - same as data_dir_root for mainnet
		"""
		if not hasattr(self, '_data_dir'):
			self._data_dir = os.path.normpath(os.path.join(*{
				'regtest': (self.data_dir_root, 'regtest', (self.regtest_user or 'none')),
				'testnet': (self.data_dir_root, 'testnet'),
				'mainnet': (self.data_dir_root,),
			}[self.network]))
		return self._data_dir

	def __init__(
			self,
			cfg          = None,
			*,
			opts_data    = None,
			init_opts    = None,
			parse_only   = False,
			parsed_opts  = None,
			need_proto   = True,
			need_amt     = True,
			caller_post_init = False,
			process_opts = False):

		# Step 1: get user-supplied configuration data from
		#           a) command line, or
		#           b) first argument to constructor;
		#         save to self._uopts:
		self._cloned = {}
		if opts_data or parsed_opts or process_opts:
			assert cfg is None, (
				'Config(): ‘cfg’ cannot be used simultaneously with ' +
				'‘opts_data’, ‘parsed_opts’ or ‘process_opts’')
			from .opts import UserOpts
			UserOpts(
				cfg          = self,
				opts_data    = opts_data,
				init_opts    = init_opts,
				parsed_opts  = parsed_opts,
				need_proto   = need_proto)
			self._uopt_src = 'cmdline'
		else:
			if cfg is None:
				self._uopts = {}
			else:
				if '_clone' in cfg:
					assert isinstance(cfg['_clone'], Config)
					self._cloned = cfg['_clone'].__dict__
					for k, v in self._cloned.items():
						if not k.startswith('_'):
							setattr(self, k, v)
					del cfg['_clone']
				self._uopts = cfg
			self._uopt_src = 'cfg'

		self._data_dir_root_override = self._cloned.pop(
			'_data_dir_root_override',
			self._uopts.pop('data_dir', None))

		if parse_only and not any(k in self._uopts for k in ['help', 'longhelp', 'usage']):
			return

		# Step 2: set cfg from user-supplied data, skipping auto opts; set type from corresponding
		#         class attribute, if it exists:
		auto_opts = tuple(self._autoset_opts) + tuple(self._auto_typeset_opts)
		for key, val in self._uopts.items():
			assert key.isascii() and key.isidentifier() and key[0] != '_', (
				f'{key!r}: malformed configuration option')
			assert key not in self._forbidden_opts, (
				f'{key!r}: forbidden configuration option')
			if key not in auto_opts:
				if hasattr(type(self), key):
					setattr(
						self,
						key,
						getattr(type(self), key) if val is None else
							conv_type(key, val, getattr(type(self), key), src=self._uopt_src))
				elif val is None:
					if hasattr(self, key):
						delattr(self, key)
				else:
					setattr(self, key, val)

		# Step 3: set cfg from environment, skipping already-set opts; save names set from environment:
		self._envopts = tuple(self._set_cfg_from_env()) if self._use_env else ()

		from .term import init_term
		init_term(self) # requires ‘hold_protect_disable’ (set from env)

		from .fileutil import check_or_create_dir
		check_or_create_dir(self.data_dir_root)

		from .util import wrap_ripemd160
		wrap_ripemd160() # ripemd160 required by mmgen_cfg_file() in _set_cfg_from_cfg_file()

		# Step 4: set cfg from cfgfile, skipping already-set opts and auto opts; save set opts and auto
		#         opts to be set:
		# requires ‘data_dir_root’, ‘test_suite_cfgtest’
		self._cfgfile_opts = self._set_cfg_from_cfg_file(self._envopts, need_proto=need_proto)

		# Step 5: set autoset opts from user-supplied data, cfgfile data, or default values, in that order:
		self._set_autoset_opts(self._cfgfile_opts.autoset)

		# Step 6: set auto typeset opts from user-supplied data or cfgfile data, in that order:
		self._set_auto_typeset_opts(self._cfgfile_opts.auto_typeset)

		# Step 7: set opts_data['sets'] opts:
		if opts_data and 'sets' in opts_data:
			self._set_opts_data_sets_opts(opts_data)

		if self.regtest or self.bob or self.alice or self.carol or gc.prog_name == f'{gc.proj_id}-regtest':
			self.network = 'regtest'
			self.regtest_user = 'bob' if self.bob else 'alice' if self.alice else 'carol' if self.carol else None
		else:
			self.network = 'testnet' if self.testnet else 'mainnet'

		self.coin = self.coin.upper()
		self.token = self.token.upper() if self.token else None

		if 'usage' in self._uopts: # requires self.coin
			import importlib
			getattr(importlib.import_module(UserOpts.help_pkg), 'usage')(self) # exits

		# self.color is finalized, so initialize color:
		if self.color: # MMGEN_DISABLE_COLOR sets this to False
			from .color import init_color
			init_color(num_colors=256 if self.force_256_color else 'auto')

		self._die_on_incompatible_opts()

		check_or_create_dir(self.data_dir)

		if self.debug and gc.prog_name != 'cmdtest.py':
			self.verbose = True
			self.quiet = False

		if self.debug_opts:
			opt_postproc_debug(self)

		from .util import Util
		self._util = Util(self)

		del self._cloned

		if hasattr(self, 'bch_cashaddr') and not hasattr(self, 'cashaddr'):
			self.cashaddr = self.bch_cashaddr

		self._lock()

		if need_proto:
			from .protocol import init_proto_from_cfg, warn_trustlevel
			# requires the default-to-none behavior, so do after the lock:
			self._proto = init_proto_from_cfg(self, need_amt=need_amt)

		if self._opts and not caller_post_init:
			self._post_init()

		# Check user-set opts without modifying them
		check_opts(self)

		if need_proto:
			warn_trustlevel(self) # do this only after proto is initialized

	def _post_init(self):
		if self.help or self.longhelp:
			from .help import print_help
			print_help(self, self._opts) # exits
		del self._opts

	def _usage(self):
		from .help import make_usage_str
		print(make_usage_str(self, caller='user'))
		sys.exit(1) # called only on bad invocation

	def _set_cfg_from_env(self):
		for name, val in ((k, v) for k, v in os.environ.items() if k.startswith('MMGEN_')):
			if name == 'MMGEN_DEBUG_ALL':
				continue
			if name in self._env_opts:
				if val: # ignore empty string values; string value of '0' or 'false' sets variable to False
					disable = name.startswith('MMGEN_DISABLE_')
					gname = name[(6, 14)[disable]:].lower()
					if gname in self._uopts: # don’t touch attr if already set by user
						continue
					if hasattr(self, gname):
						setattr(
							self,
							gname,
							conv_type(name, val, getattr(self, gname), src='env', invert_bool=disable))
						yield gname
					else:
						raise ValueError(f'Name {gname!r} not present in globals')
			else:
				raise ValueError(f'{name!r} is not a valid MMGen environment variable')

	def _set_cfg_from_cfg_file(self, env_cfg, *, need_proto):

		_ret = namedtuple('cfgfile_opts', ['non_auto', 'autoset', 'auto_typeset'])

		if not self._use_cfg_file:
			return _ret((), {}, {})

		# check for changes in system template file (term must be initialized)
		from .cfgfile import mmgen_cfg_file
		mmgen_cfg_file(self, 'sample')

		ucfg = mmgen_cfg_file(self, 'usr')

		self._cfgfile_fn = ucfg.fn

		if need_proto:
			from .protocol import init_proto

		autoset_opts = {}
		auto_typeset_opts = {}
		non_auto_opts = []
		already_set = tuple(self._uopts) + env_cfg

		def set_opt(d, obj, name, refval):
			val = ucfg.parse_value(d.value, refval)
			if not val:
				die('CfgFileParseError', f'Parse error in file {ucfg.fn!r}, line {d.lineno}')
			val_conv = conv_type(name, val, refval, src=ucfg.fn)
			setattr(obj, name, val_conv)
			non_auto_opts.append(name)

		for d in ucfg.get_lines():
			if d.name in self._cfg_file_opts:
				if not d.name in already_set:
					set_opt(d, self, d.name, getattr(self, d.name))
			elif d.name in self._autoset_opts:
				autoset_opts[d.name] = d.value
			elif d.name in self._auto_typeset_opts:
				auto_typeset_opts[d.name] = d.value
			elif any(d.name.startswith(coin + '_') for coin in gc.rpc_coins):
				if need_proto and not d.name in already_set:
					try:
						refval = init_proto(self, d.name.split('_', 1)[0]).get_opt_clsval(self, d.name)
					except AttributeError:
						die('CfgFileParseError', f'{d.name!r}: unrecognized option in {ucfg.fn!r}, line {d.lineno}')
					set_opt(d, self, d.name, refval)
			else:
				die('CfgFileParseError', f'{d.name!r}: unrecognized option in {ucfg.fn!r}, line {d.lineno}')

		return _ret(tuple(non_auto_opts), autoset_opts, auto_typeset_opts)

	def _set_autoset_opts(self, cfgfile_autoset_opts):

		def get_autoset_opt(key, val, src):

			def die_on_err(desc):
				from .util import fmt_list
				die(
					'UserOptError',
					'{a!r}: invalid {b} (not {c}: {d})'.format(
						a = val,
						b = {
							'cmdline': f'parameter for option --{key.replace("_", "-")}',
							'cfgfile': f'value for cfg file option {key!r}'
						}[src],
						c = desc,
						d = fmt_list(data.choices)))

			class opt_type:

				def nocase_str():
					if val.lower() in data.choices:
						return val.lower()
					else:
						die_on_err('one of')

				def nocase_pfx():
					cs = [s for s in data.choices if s.startswith(val.lower())]
					if len(cs) == 1:
						return cs[0]
					else:
						die_on_err('unique substring of')

			data = self._autoset_opts[key]

			return getattr(opt_type, data.type)()

		# Check autoset opts, setting if unset
		for key in self._autoset_opts:

			if key in self._cloned:
				continue

			assert not hasattr(self, key), f'autoset opt {key!r} is already set, but it shouldn’t be!'

			if key in self._uopts:
				val, src = (self._uopts[key], 'cmdline')
			elif key in cfgfile_autoset_opts:
				val, src = (cfgfile_autoset_opts[key], 'cfgfile')
			else:
				val = None

			if val is None:
				if key not in self._dfl_none_autoset_opts:
					setattr(self, key, self._autoset_opts[key].choices[0])
			else:
				setattr(self, key, get_autoset_opt(key, val, src=src))

	def _set_auto_typeset_opts(self, cfgfile_auto_typeset_opts):

		def do_set(key, val, ref_type):
			assert not hasattr(self, key), f'{key!r} is in cfg!'
			setattr(self, key, None if val is None else ref_type(val))

		for key, ref_type in self._auto_typeset_opts.items():
			if key in self._uopts:
				do_set(key, self._uopts[key], ref_type)
			elif key in cfgfile_auto_typeset_opts:
				do_set(key, cfgfile_auto_typeset_opts[key], ref_type)

	def _set_opts_data_sets_opts(self, opts_data):
		for a_opt, a_val, b_opt, b_val in opts_data['sets']:
			if (usr_a_val := getattr(self, a_opt, None)) not in (None, False):
				if a_val == bool or usr_a_val == a_val:
					if ((usr_b_val := getattr(self, b_opt, None)) in (None, False)) or usr_b_val == b_val:
						setattr(self, b_opt, b_val)
					else:
						die(1, 'Option --{}={} conflicts with option --{}={}\n'.format(
							b_opt.replace('_', '-'),
							usr_b_val,
							a_opt.replace('_', '-'),
							usr_a_val))

	def _die_on_incompatible_opts(self):
		for group in self._incompatible_opts:
			bad = [k for k in self.__dict__ if k in group and getattr(self, k) is not None]
			if len(bad) > 1:
				die(1, 'Conflicting options: {}'.format(', '.join(map(fmt_opt, bad))))

	def _set_quiet(self, val):
		from .util import Util
		self.__dict__['quiet'] = val
		self.__dict__['_util'] = Util(self) # qmsg, qmsg_r

def check_opts(cfg): # Raises exception if any check fails

	from .util import is_int, Msg

	def get_desc(desc_pfx=''):
		return (
			(desc_pfx + ' ' if desc_pfx else '')
			+ (
				f'parameter for command-line option {fmt_opt(name)!r}'
					if name in cfg._uopts and cfg._uopt_src == 'cmdline' else
				f'value for configuration option {name!r}'
			)
			+ (' from environment' if name in cfg._envopts else '')
			+ (f' in {cfg._cfgfile_fn!r}' if name in cfg._cfgfile_opts.non_auto else '')
		)

	def display_opt(name, val='', *, beg='For selected', end=':\n'):
		from .util import msg_r
		msg_r('{} option {!r}{}'.format(
			beg,
			f'{fmt_opt(name)}={val}' if val else fmt_opt(name),
			end))

	def opt_compares(val, op_str, target):
		import operator
		if not {
			'<':  operator.lt,
			'<=': operator.le,
			'>':  operator.gt,
			'>=': operator.ge,
			'=':  operator.eq,
		}[op_str](val, target):
			die('UserOptError', f'{val}: invalid {get_desc()} (not {op_str} {target})')

	def opt_is_int(val, *, desc_pfx=''):
		if not is_int(val):
			die('UserOptError', f'{val!r}: invalid {get_desc(desc_pfx)} (not an integer)')

	def opt_is_in_list(val, tlist, *, desc_pfx=''):
		if val not in tlist:
			q, sep = (('', ','), ("'", "','"))[isinstance(tlist[0], str)]
			die('UserOptError', '{q}{v}{q}: invalid {w}\nValid choices: {q}{o}{q}'.format(
				v = val,
				w = get_desc(desc_pfx),
				q = q,
				o = sep.join(map(str, sorted(tlist)))))

	def opt_unrecognized():
		die('UserOptError', f'{val!r}: unrecognized {get_desc()}')

	class check_funcs:

		def in_fmt():
			from .wallet import get_wallet_data
			wd = get_wallet_data(fmt_code=val)
			if not wd:
				opt_unrecognized()
			if name == 'out_fmt':
				p = 'hidden_incog_output_params'

				if wd.type == 'incog_hidden' and not getattr(cfg, p):
					die('UserOptError',
						'Hidden incog format output requested.  ' +
						f'You must supply a file and offset with the {fmt_opt(p)!r} option')

				if wd.base_type == 'incog_base' and cfg.old_incog_fmt:
					display_opt(name, val, beg='Selected', end=' ')
					display_opt('old_incog_fmt', beg='conflicts with', end=':\n')
					die('UserOptError', 'Export to old incog wallet format unsupported')
				elif wd.type == 'brain':
					die('UserOptError', 'Output to brainwallet format unsupported')

		out_fmt = in_fmt

		def hidden_incog_input_params():
			a = val.rsplit(',', 1) # permit comma in filename
			if len(a) != 2:
				display_opt(name, val)
				die('UserOptError', 'Option requires two comma-separated arguments')

			fn, offset = a
			opt_is_int(offset)

			from .fileutil import check_infile, check_outdir, check_outfile
			if name == 'hidden_incog_input_params':
				check_infile(fn, blkdev_ok=True)
				key2 = 'in_fmt'
			else:
				try:
					os.stat(fn)
				except:
					b = os.path.dirname(fn)
					if b:
						check_outdir(b)
				else:
					check_outfile(fn, blkdev_ok=True)
				key2 = 'out_fmt'

			if hasattr(cfg, key2):
				val2 = getattr(cfg, key2)
				from .wallet import get_wallet_data
				wd = get_wallet_data(wtype='incog_hidden')
				if val2 and val2 not in wd.fmt_codes:
					die('UserOptError',
						f'Option {fmt_opt(name)} conflicts with option {fmt_opt(key2)}={val2}')

		hidden_incog_output_params = hidden_incog_input_params

		def subseeds():
			from .subseed import SubSeedIdxRange
			opt_compares(val, '>=', SubSeedIdxRange.min_idx)
			opt_compares(val, '<=', SubSeedIdxRange.max_idx)

		def seed_len():
			from .seed import Seed
			opt_is_in_list(int(val), Seed.lens)

		def hash_preset():
			from .crypto import Crypto
			opt_is_in_list(val, list(Crypto.hash_presets.keys()))

		def brain_params():
			a = val.split(',')
			if len(a) != 2:
				display_opt(name, val)
				die('UserOptError', 'Option requires two comma-separated arguments')

			opt_is_int(a[0], desc_pfx='seed length')
			from .seed import Seed
			opt_is_in_list(int(a[0]), Seed.lens, desc_pfx='seed length')

			from .crypto import Crypto
			opt_is_in_list(a[1], list(Crypto.hash_presets.keys()), desc_pfx='hash preset')

		def usr_randchars():
			if val != 0:
				opt_compares(val, '>=', cfg.min_urandchars)
				opt_compares(val, '<=', cfg.max_urandchars)

		def tx_confs():
			opt_is_int(val)
			opt_compares(int(val), '>=', 1)

		def vsize_adj():
			from .util import ymsg
			ymsg(f'Adjusting transaction vsize by a factor of {val:1.2f}')

		def daemon_id():
			from .daemon import CoinDaemon
			opt_is_in_list(val, CoinDaemon.all_daemon_ids())

		def locktime():
			opt_is_int(val)
			opt_compares(int(val), '>', 0)

		def columns():
			opt_compares(val, '>', 10)

	# TODO: add checks for token, rbf, tx_fee
	check_funcs_names = tuple(check_funcs.__dict__)
	for name in tuple(cfg._uopts) + cfg._envopts + cfg._cfgfile_opts.non_auto:
		val = getattr(cfg, name)
		if name in cfg._infile_opts:
			from .fileutil import check_infile
			check_infile(val) # file exists and is readable - dies on error
		elif name == 'outdir':
			from .fileutil import check_outdir
			check_outdir(val) # dies on error
		elif name in check_funcs_names:
			getattr(check_funcs, name)()
		elif cfg.debug:
			Msg(f'check_opts(): No test for config opt {name!r}')

def fmt_opt(o):
	return '--' + o.replace('_', '-')

def opt_postproc_debug(cfg):
	none_opts = [k for k in dir(cfg) if k[:2] != '__' and getattr(cfg, k) is None]
	from .util import Msg
	Msg('\n    Configuration opts:')
	for e in [d for d in dir(cfg) if d[:2] != '__']:
		Msg(f'        {e:<20}: {getattr(cfg, e)}')
	Msg("    Configuration opts set to 'None':")
	Msg('        {}\n'.format('\n        '.join(none_opts)))
	Msg('\n=== end opts.py debug ===\n')

def conv_type(name, val, refval, *, src, invert_bool=False):

	def do_fail():
		desc = {
			'cmdline': 'command-line',
			'cfg':     'Config',
			'env':     'environment var',
		}
		die(1, '{a!r}: invalid value for {b} option {c!r}{d} (must be of type {e!r})'.format(
			a = val,
			b = desc.get(src, 'config file'),
			c = fmt_opt(name) if src == 'cmdline' else name,
			d = '' if src in ('cmdline', 'cfg', 'env') else f' in {src!r}',
			e = type(refval).__name__))

	# refval is None = boolean opt with no cmdline parameter
	if type(refval) is bool or refval is None:
		v = str(val).lower()
		ret = (
			True  if v in ('true', 'yes', '1', 'on') else
			False if v in ('false', 'no', 'none', '0', 'off', '') else
			None
		)
		return do_fail() if ret is None else (not ret) if invert_bool else ret
	elif isinstance(refval, (list, tuple)):
		if src == 'cmdline':
			return type(refval)(val.split(','))
		else:
			assert isinstance(val, (list, tuple)), f'{val}: not a list or tuple'
			return type(refval)(val)
	else:
		try:
			return type(refval)(not val if invert_bool else val)
		except:
			do_fail()



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
util2: Less frequently-used variables, classes and utility functions for the MMGen suite
"""

import sys, re, time
from .util import msg, suf, hexdigits, die

def die_wait(delay, ev=0, s=''):
	assert isinstance(delay, int)
	assert isinstance(ev, int)
	if s:
		msg(s)
	time.sleep(delay)
	sys.exit(ev)

def die_pause(ev=0, s=''):
	assert isinstance(ev, int)
	if s:
		msg(s)
	input('Press ENTER to exit')
	sys.exit(ev)

def cffi_override_fixup():
	from cffi import FFI
	class FFI_override:
		def cdef(self, csource, *, override=False, packed=False, pack=None):
			self._cdef(csource, override=True, packed=packed, pack=pack)
	FFI.cdef = FFI_override.cdef

# monkey-patch function: makes modules pycryptodome and pycryptodomex available to packages that
# expect them (monero-python, eth-keys), regardless of which one is installed on system
def load_cryptodome(called=[]):
	if not called:
		cffi_override_fixup()
		try:
			import Crypto # Crypto == pycryptodome
		except ImportError:
			try:
				import Cryptodome # Crypto == pycryptodome
			except ImportError:
				die(2, 'Unable to import the ‘pycryptodome’ or ‘pycryptodomex’ package')
			else:
				sys.modules['Crypto'] = Cryptodome # Crypto == pycryptodome
		else:
			sys.modules['Cryptodome'] = Crypto # Cryptodome == pycryptodomex
		called.append(True)

# called with no arguments by proto.eth.tx.transaction:
def get_keccak(cfg=None, cached_ret=[]):

	if not cached_ret:
		if cfg and cfg.use_internal_keccak_module:
			cfg._util.qmsg('Using internal keccak module by user request')
			from .contrib.keccak import keccak_256
		else:
			load_cryptodome()
			from Crypto.Hash import keccak
			def keccak_256(data):
				return keccak.new(data=data, digest_bytes=32)
		cached_ret.append(keccak_256)

	return cached_ret[0]

# From 'man dd':
# c=1, w=2, b=512, kB=1000, K=1024, MB=1000*1000, M=1024*1024,
# GB=1000*1000*1000, G=1024*1024*1024, and so on for T, P, E, Z, Y.
bytespec_map = (
	('c',  1),
	('w',  2),
	('b',  512),
	('kB', 1000),
	('K',  1024),
	('MB', 1000000),
	('M',  1048576),
	('GB', 1000000000),
	('G',  1073741824),
	('TB', 1000000000000),
	('T',  1099511627776),
	('PB', 1000000000000000),
	('P',  1125899906842624),
	('EB', 1000000000000000000),
	('E',  1152921504606846976),
)

def int2bytespec(n, spec, fmt, *, print_sym=True, strip=False, add_space=False):

	def spec2int(spec):
		for k, v in bytespec_map:
			if k == spec:
				return v
		die(1, f'{spec!r}: unrecognized bytespec')

	ret = f'{n/spec2int(spec):{fmt}f}'
	if strip:
		ret = ret.rstrip('0')
		return (
			ret
			+ ('0' if ret.endswith('.') else '')
			+ ((' ' if add_space else '') + spec if print_sym else ''))
	else:
		return (
			ret
			+ ((' ' if add_space else '') + spec if print_sym else ''))

def parse_bytespec(nbytes):
	m = re.match(r'([0123456789.]+)(.*)', nbytes)
	if m:
		if m.group(2):
			for k, v in bytespec_map:
				if k == m.group(2):
					from decimal import Decimal
					return int(Decimal(m.group(1)) * v)
			msg("Valid byte specifiers: '{}'".format("' '".join([i[0] for i in bytespec_map])))
		elif '.' in nbytes:
			raise ValueError('fractional bytes not allowed')
		else:
			return int(nbytes)

	die(1, f'{nbytes!r}: invalid byte specifier')

def format_elapsed_days_hr(t, *, now=None, cached={}):
	e = int((now or time.time()) - t)
	if not e in cached:
		days = abs(e) // 86400
		cached[e] = f'{days} day{suf(days)} ' + ('ago' if e > 0 else 'in the future')
	return cached[e]

def format_elapsed_hr(
		t,
		*,
		now        = None,
		cached     = {},
		rel_now    = True,
		show_secs  = False,
		future_msg = 'in the future'):
	e = int((now or time.time()) - t)
	key = f'{e}:{rel_now}:{show_secs}'
	if not key in cached:
		def add_suffix():
			return (
				((' ago'           if rel_now else '') if e > 0 else
				(f' {future_msg}' if rel_now else ' (negative elapsed)'))
					if (abs_e if show_secs else abs_e // 60) else
				('just now' if rel_now else ('0 ' + ('seconds' if show_secs else 'minutes')))
			)
		abs_e = abs(e)
		data = (
			('day',    abs_e // 86400),
			('hour',   abs_e // 3600 % 24),
			('minute', abs_e // 60 % 60),
			('second', abs_e % 60),
		) if show_secs else (
			('day',    abs_e // 86400),
			('hour',   abs_e // 3600 % 24),
			('minute', abs_e // 60 % 60),
		)
		cached[key] = ' '.join(f'{n} {desc}{suf(n)}' for desc, n in data if n) + add_suffix()
	return cached[key]

def pretty_format(s, *, width=80, pfx=''):
	out = []
	while s:
		if len(s) <= width:
			out.append(s)
			break
		i = s[:width].rfind(' ')
		out.append(s[:i])
		s = s[i+1:]
	return pfx + ('\n'+pfx).join(out)

def block_format(data, *, gw=2, cols=8, line_nums=None, data_is_hex=False):
	assert line_nums in (None, 'hex', 'dec'), "'line_nums' must be one of None, 'hex' or 'dec'"
	ln_fs = '{:06x}: ' if line_nums == 'hex' else '{:06}: '
	bytes_per_chunk = gw
	if data_is_hex:
		gw *= 2
	nchunks = len(data)//gw + bool(len(data)%gw)
	return ''.join(
		('' if (line_nums is None or i % cols) else ln_fs.format(i*bytes_per_chunk))
		+ data[i*gw:i*gw+gw]
		+ (' ' if (not cols or (i+1) % cols) else '\n')
			for i in range(nchunks)
	).rstrip() + '\n'

def pretty_hexdump(data, *, gw=2, cols=8, line_nums=None):
	return block_format(data.hex(), gw=gw, cols=cols, line_nums=line_nums, data_is_hex=True)

def decode_pretty_hexdump(data):
	pat = re.compile(fr'^[{hexdigits}]+:\s+')
	lines = [pat.sub('', line) for line in data.splitlines()]
	try:
		return bytes.fromhex(''.join((''.join(lines).split())))
	except:
		msg('Data not in hexdump format')
		return False

def cliargs_convert(args):

	# return str instead of float for input into JSON-RPC
	def float_parser(n):
		return n

	import json
	def gen():
		for arg in args:
			try:
				yield json.loads(arg, parse_float=float_parser) # list, dict, bool, int, null, float
			except json.decoder.JSONDecodeError:
				yield arg # arbitrary string

	return tuple(gen())

def port_in_use(port):
	import socket
	try:
		socket.create_connection(('localhost', port)).close()
	except:
		return False
	else:
		return True

class ExpInt(int):
	'encode or parse an integer in exponential notation with specified precision'

	max_prec = 10

	def __new__(cls, spec, *, prec):
		assert 0 < prec < cls.max_prec
		cls.prec = prec

		from .util import is_int
		if is_int(spec):
			return int.__new__(cls, spec)
		else:
			assert isinstance(spec, str), f'ExpInt: {spec!r}: not a string!'
			assert len(spec) >= 3, f'ExpInt: {spec!r}: invalid specifier'
			val, exp = spec.split('e')
			assert is_int(val) and is_int(exp)
			return int.__new__(cls, val + '0' * int(exp))

	@property
	def trunc(self):
		s = str(self)
		return int(s[:self.prec] + '0' * (len(s) - self.prec))

	@property
	def enc(self):
		s = str(self)
		s_len = len(s)
		digits = s[:min(s_len, self.prec)].rstrip('0')
		ret = '{}e{}'.format(digits, s_len - len(digits))
		return ret if len(ret) < s_len else s



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
baseconv: base conversion class for the MMGen suite
"""

from collections import namedtuple

from .util import die

def is_b58_str(s):
	return set(list(s)) <= set(baseconv('b58').digits)

def is_b32_str(s):
	return set(list(s)) <= set(baseconv('b32').digits)

def is_mmgen_mnemonic(s):
	try:
		baseconv('mmgen').tobytes(s.split(), pad='seed')
		return True
	except:
		return False

class baseconv:
	mn_base = 1626
	dt = namedtuple('desc_tuple', ['short', 'long'])
	constants = {
	'desc': {
		'b58':   dt('base58',             'base58-encoded data'),
		'b32':   dt('MMGen base32',       'MMGen base32-encoded data created using simple base conversion'),
		'b16':   dt('hexadecimal string', 'base16 (hexadecimal) string data'),
		'b10':   dt('base10 string',      'base10 (decimal) string data'),
		'b8':    dt('base8 string',       'base8 (octal) string data'),
		'b6d':   dt('base6d (die roll)',  'base6 data using the digits from one to six'),
		'mmgen': dt('MMGen native mnemonic',
		'MMGen native mnemonic seed phrase created using old Electrum wordlist and simple base conversion'),
	},
	# https://en.wikipedia.org/wiki/Base32#RFC_4648_Base32_alphabet
	# https://tools.ietf.org/html/rfc4648
	'digits': {
		'b58': tuple('123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz'),
		'b32': tuple('ABCDEFGHIJKLMNOPQRSTUVWXYZ234567'), # RFC 4648 alphabet
		'b16': tuple('0123456789abcdef'),
		'b10': tuple('0123456789'),
		'b8':  tuple('01234567'),
		'b6d': tuple('123456'),
	},
	'wl_chksum': {
		'mmgen':  '5ca31424',
#		'tirosh': '48f05e1f', # tirosh truncated to mn_base
#		'tirosh1633': '1a5faeff' # tirosh list is 1633 words long!
	},
	'seedlen_map': {
		'b58':   {16:22, 24:33, 32:44},
		'b6d':   {16:50, 24:75, 32:100},
		'mmgen': {16:12, 24:18, 32:24},
	},
	'seedlen_map_rev': {
		'b58':   {22:16, 33:24, 44:32},
		'b6d':   {50:16, 75:24, 100:32},
		'mmgen': {12:16, 18:24, 24:32},
	}
	}

	def __init__(self, wl_id):

		if wl_id == 'mmgen':
			from .wordlist.electrum import words
			self.constants['digits'][wl_id] = words
		elif wl_id not in self.constants['digits']:
			raise ValueError(f'{wl_id}: unrecognized mnemonic ID')

		for k, v in self.constants.items():
			if wl_id in v:
				setattr(self, k, v[wl_id])

		self.wl_id = wl_id

	def get_wordlist(self):
		return self.digits

	def get_wordlist_chksum(self):
		from hashlib import sha256
		return sha256(' '.join(self.digits).encode()).hexdigest()[:8]

	def check_wordlist(self, cfg):

		wl = self.digits
		ret = f'Wordlist: {self.wl_id}\nLength: {len(wl)} words'
		new_chksum = self.get_wordlist_chksum()

		cfg._util.compare_chksums(new_chksum, 'generated', self.wl_chksum, 'saved', die_on_fail=True)

		if tuple(sorted(wl)) == wl:
			return ret + '\nList is sorted'
		else:
			die(3, 'ERROR: List is not sorted!')

	@staticmethod
	def get_pad(pad, /, seed_pad_func):
		"""
		'pad' argument to baseconv conversion methods must be either None, 'seed' or an integer.
		If None, output of minimum (but never zero) length will be produced.
		If 'seed', output length will be mapped from input length using data in seedlen_map.
		If an integer, the string, hex string or byte output will be padded to this length.
		"""
		if pad is None:
			return 0
		elif type(pad) is int:
			return pad
		elif pad == 'seed':
			return seed_pad_func()
		else:
			die('BaseConversionPadError', f"{pad!r}: illegal value for 'pad' (must be None, 'seed' or int)")

	def tohex(self, words_arg, /, *, pad=None):
		"convert string or list data of instance base to a hexadecimal string"
		return self.tobytes(words_arg, pad=pad//2 if type(pad) is int else pad).hex()

	def tobytes(self, words_arg, /, *, pad=None):
		"convert string or list data of instance base to byte string"

		words = words_arg if isinstance(words_arg, (list, tuple)) else tuple(words_arg.strip())
		desc = self.desc.short

		if len(words) == 0:
			die('BaseConversionError', f'empty {desc} data')

		def get_seed_pad():
			assert hasattr(self, 'seedlen_map_rev'), f'seed padding not supported for base {self.wl_id!r}'
			d = self.seedlen_map_rev
			if not len(words) in d:
				die('BaseConversionError',
					f'{len(words)}: invalid length for seed-padded {desc} data in base conversion')
			return d[len(words)]

		pad_val = max(self.get_pad(pad, get_seed_pad), 1)
		wl = self.digits
		base = len(wl)

		if not set(words) <= set(wl):
			die('BaseConversionError',
				('seed data' if pad == 'seed' else f'{words_arg!r}:') +
				f' not in {desc} format')

		ret = sum(wl.index(words[::-1][i])*(base**i) for i in range(len(words)))
		bl = ret.bit_length()
		return ret.to_bytes(max(pad_val, bl//8+bool(bl%8)), 'big')

	def fromhex(self, hexstr, /, *, pad=None, tostr=False):
		"convert a hexadecimal string to a list or string data of instance base"

		from .util import is_hex_str
		if not is_hex_str(hexstr):
			die('HexadecimalStringError',
				('seed data' if pad == 'seed' else f'{hexstr!r}:') +
				' not a hexadecimal string')

		return self.frombytes(bytes.fromhex(hexstr), pad=pad, tostr=tostr)

	def frombytes(self, bytestr, /, *, pad=None, tostr=False):
		"convert byte string to list or string data of instance base"

		if not bytestr:
			die('BaseConversionError', 'empty data not allowed in base conversion')

		def get_seed_pad():
			assert hasattr(self, 'seedlen_map'), f'seed padding not supported for base {self.wl_id!r}'
			d = self.seedlen_map
			if not len(bytestr) in d:
				die('SeedLengthError',
					f'{len(bytestr)}: invalid byte length for seed data in seed-padded base conversion')
			return d[len(bytestr)]

		pad = max(self.get_pad(pad, get_seed_pad), 1)
		wl = self.digits

		def gen():
			num = int.from_bytes(bytestr, 'big')
			base = len(wl)
			while num:
				yield num % base
				num //= base

		ret = list(gen())
		o = [wl[n] for n in [0] * (pad-len(ret)) + ret[::-1]]

		return (' ' if self.wl_id == 'mmgen' else '').join(o) if tostr else o



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
xmrwallet.file.outputs: Monero wallet outputs file class for the MMGen Suite
"""

import re
from collections import namedtuple
from pathlib import Path

from ...util import die, suf
from ...tx.util import get_autosign_obj

from . import MoneroMMGenFile

class MoneroWalletOutputsFile:

	class Base(MoneroMMGenFile):

		desc = 'wallet outputs'
		data_label = 'MoneroMMGenWalletOutputsFile'
		base_chksum_fields = {'seed_id', 'wallet_index', 'outputs_data_hex'}
		full_chksum_fields = {'seed_id', 'wallet_index', 'outputs_data_hex', 'signed_key_images'}
		fn_fs = '{a}-outputs-{b}.{c}'
		ext_offset = 25 # len('-outputs-') + len(chksum) ({b})
		chksum_nchars = 16
		data_tuple = namedtuple('wallet_outputs_data', [
			'seed_id',
			'wallet_index',
			'outputs_data_hex',
			'signed_key_images',
			'sign',
			'imported',
		])

		def __init__(self, cfg):
			self.name = type(self).__name__
			self.cfg = cfg

		def write(self, *, add_suf='', quiet=False):
			from ...fileutil import write_data_to_file
			write_data_to_file(
				cfg               = self.cfg,
				outfile           = str(self.get_outfile(self.cfg, self.wallet_fn)) + add_suf,
				data              = self.make_wrapped_data(self.data._asdict()),
				desc              = self.desc,
				ask_overwrite     = False,
				quiet             = quiet,
				ignore_opt_outdir = True)

		def get_outfile(self, cfg, wallet_fn):
			return (
				get_autosign_obj(cfg).xmr_outputs_dir if cfg.autosign else
				wallet_fn.parent) / self.fn_fs.format(
					a = wallet_fn.name,
					b = self.base_chksum,
					c = self.ext,
				)

		def get_wallet_fn(self, fn):
			assert fn.name.endswith(f'.{self.ext}'), (
				f'{self.name}: filename does not end with {"."+self.ext!r}'
			)
			return fn.parent / fn.name[:-(len(self.ext)+self.ext_offset+1)]

		def get_info(self, *, indent=''):
			if self.data.signed_key_images is not None:
				data = self.data.signed_key_images or []
				return f'{indent}{self.wallet_fn.name}: {len(data)} signed key image{suf(data)}'
			else:
				return f'{indent}{self.wallet_fn.name}: no key images'

	class New(Base):
		ext = 'raw'

		def __init__(self, parent, wallet_fn, data, *, wallet_idx=None, sign=False):
			super().__init__(parent.cfg)
			self.wallet_fn = wallet_fn
			init_data = dict.fromkeys(self.data_tuple._fields)
			init_data.update({
				'seed_id':      parent.kal.al_id.sid,
				'wallet_index': wallet_idx or parent.get_idx_from_fn(wallet_fn),
			})
			if sign:
				init_data['sign'] = sign
			init_data.update({k:v for k, v in data.items() if k in init_data})
			self.data = self.data_tuple(**init_data)

	class Completed(New):

		def __init__(self, parent, *, fn=None, wallet_fn=None):
			def check_equal(desc, a, b):
				assert a == b, f'{desc} mismatch: {a} (from file) != {b} (from filename)'
			fn = fn or self.get_outfile(parent.cfg, wallet_fn)
			wallet_fn = wallet_fn or self.get_wallet_fn(fn)
			d_wrap = self.extract_data_from_file(parent.cfg, fn)
			data = d_wrap['data']
			check_equal('Seed ID', data['seed_id'], parent.kal.al_id.sid)
			wallet_idx = parent.get_idx_from_fn(wallet_fn)
			check_equal('Wallet index', data['wallet_index'], wallet_idx)
			super().__init__(
				parent     = parent,
				wallet_fn  = wallet_fn,
				data       = data,
				wallet_idx = wallet_idx,
			)
			self.check_checksums(d_wrap)

		@classmethod
		def find_fn_from_wallet_fn(cls, cfg, wallet_fn, *, ret_on_no_match=False):
			path = get_autosign_obj(cfg).xmr_outputs_dir or Path()
			pat = cls.fn_fs.format(
				a = wallet_fn.name,
				b = f'[0-9a-f]{{{cls.chksum_nchars}}}\\',
				c = cls.ext,
			)
			matches = [f for f in path.iterdir() if re.match(pat, f.name)]
			if not matches and ret_on_no_match:
				return None
			if not matches or len(matches) > 1:
				die(2, "{a} matching pattern {b!r} found in '{c}'!".format(
					a = 'No files' if not matches else 'More than one file',
					b = pat,
					c = path
				))
			return matches[0]

	class Unsigned(Completed):
		pass

	class SignedNew(New):
		desc = 'signed key images'
		ext = 'sig'

	class Signed(Completed, SignedNew):
		pass

class MoneroWalletDumpFile:

	class Base:
		desc = 'Monero wallet dump'
		data_label = 'MoneroMMGenWalletDumpFile'
		base_chksum_fields = {'seed_id', 'wallet_index', 'wallet_metadata'}
		full_chksum_fields = None
		ext = 'dump'
		ext_offset = 0
		data_tuple = namedtuple('wallet_dump_data', [
			'seed_id',
			'wallet_index',
			'wallet_metadata',
		])
		def get_outfile(self, cfg, wallet_fn):
			return wallet_fn.parent / f'{wallet_fn.name}.{self.ext}'

	class New(Base, MoneroWalletOutputsFile.New):
		pass

	class Completed(Base, MoneroWalletOutputsFile.Completed):
		pass



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
wallet.base: wallet base class
"""

import os

from ..util import msg, die
from ..color import orange
from ..objmethods import MMGenObject
from . import wallet_data, get_wallet_cls

class WalletMeta(type):

	def __init__(cls, name, bases, namespace):
		t = cls.__module__.rsplit('.', maxsplit=1)[-1]
		if t in wallet_data:
			for k, v in wallet_data[t]._asdict().items():
				setattr(cls, k, v)

class wallet(MMGenObject, metaclass=WalletMeta):

	desc = 'MMGen seed source'
	file_mode = 'text'
	filename_api = True
	stdin_ok = False
	ask_tty = True
	no_tty  = False
	op = None

	class WalletData(MMGenObject):
		pass

	def __init__(self, *,
		in_data       = None,
		passwd_file   = None):

		self.passwd_file = None if passwd_file is False else (passwd_file or self.cfg.passwd_file)
		self.ssdata = self.WalletData()
		self.msg = {}
		self.in_data = in_data

		for c in reversed(self.__class__.__mro__):
			if hasattr(c, '_msg'):
				self.msg.update(c._msg)

		if hasattr(self, 'seed'):
			self._encrypt()
			return
		elif hasattr(self, 'infile') or self.in_data or not self.cfg.stdin_tty:
			self._deformat_once()
			self._decrypt_retry()
		else:
			if not self.stdin_ok:
				die(1, f'Reading from standard input not supported for {self.desc} format')
			self._deformat_retry()
			self._decrypt_retry()

		self.cfg._util.qmsg('Valid {a} for Seed ID {b}{c}{d}'.format(
			a = self.desc,
			b = self.seed.sid.hl(),
			c = f' (seed length {self.seed.bitlen})' if self.seed.bitlen != 256 else '',
			d = '' if not hasattr(self, 'mnemonic') or self.mnemonic.has_chksum else
				orange(' [mnemonic format has no checksum]')
		))

	def _get_data(self):
		if hasattr(self, 'infile'):
			from ..fileutil import get_data_from_file
			self.fmt_data = get_data_from_file(
				self.cfg,
				self.infile.name,
				desc   = self.desc,
				binary = self.file_mode=='binary')
		elif self.in_data:
			self.fmt_data = self.in_data
		else:
			self.fmt_data = self._get_data_from_user(self.desc)

	def _get_data_from_user(self, desc):
		from ..ui import get_data_from_user
		return get_data_from_user(self.cfg, desc=desc)

	def _deformat_once(self):
		self._get_data()
		if not self._deformat():
			die(2, 'Invalid format for input data')

	def _deformat_retry(self):
		while True:
			self._get_data()
			if self._deformat():
				break
			msg('Trying again...')

	@staticmethod
	def ext_to_cls(ext, proto):
		return get_wallet_cls(ext=ext)

	def get_fmt_data(self):
		self._format()
		return self.fmt_data

	def write_to_file(self, *, outdir='', desc=''):
		self._format()
		kwargs = {
			'desc':     desc or self.desc,
			'ask_tty':  self.ask_tty,
			'no_tty':   self.no_tty,
			'binary':   self.file_mode == 'binary'
		}

		if outdir:
			# write_data_to_file(): outfile with absolute path overrides self.cfg.outdir
			of = os.path.abspath(os.path.join(outdir, self._filename()))

		from ..fileutil import write_data_to_file
		write_data_to_file(
			self.cfg,
			of if outdir else self._filename(),
			self.fmt_data,
			**kwargs)

	def check_usr_seed_len(self, *, bitlen=None):
		chk = bitlen or self.seed.bitlen
		if self.cfg.seed_len and self.cfg.seed_len != chk:
			die(1, f'ERROR: requested seed length ({self.cfg.seed_len}) doesn’t match seed length of source ({chk})')



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
xmrwallet.file: Monero file base class for the MMGen Suite
"""

import json
from ...util import make_chksum_N
from ...fileutil import get_data_from_file
from ...rpc.util import json_encoder

class MoneroMMGenFile:

	silent_load = False

	def make_chksum(self, *, keys=None):
		res = json.dumps(
			dict((k, v) for k, v in self.data._asdict().items() if (not keys or k in keys)),
			cls = json_encoder
		)
		return make_chksum_N(res, rounds=1, nchars=self.chksum_nchars, upper=False)

	@property
	def base_chksum(self):
		return self.make_chksum(keys=self.base_chksum_fields)

	@property
	def full_chksum(self):
		return self.make_chksum(keys=self.full_chksum_fields) if self.full_chksum_fields else None

	def check_checksums(self, d_wrap):
		for k in ('base_chksum', 'full_chksum'):
			a = getattr(self, k)
			if a is not None:
				b = d_wrap[k]
				assert a == b, f'{k} mismatch: {a} != {b}'

	def make_wrapped_data(self, in_data):
		out = {
			'base_chksum': self.base_chksum,
			'full_chksum': self.full_chksum,
			'data': in_data,
		} if self.full_chksum else {
			'base_chksum': self.base_chksum,
			'data': in_data,
		}
		return json.dumps(
			{ self.data_label: out },
			cls = json_encoder,
			indent = 2,
		)

	def extract_data_from_file(self, cfg, fn):
		return json.loads(
			get_data_from_file(cfg, str(fn), desc=self.desc, silent=self.silent_load)
		)[self.data_label]



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
mmgen-txdo: Create, sign and send an online MMGen transaction
"""

from .cfg import gc, Config
from .util import die, fmt_list, async_run
from .subseed import SubSeedIdxRange

target = gc.prog_name.split('-')[1].removesuffix('do')

opts_data = {
	'filter_codes': {
		'tx':     ['-', 't'],
		'swaptx': ['-', 's'],
	}[target],
	'sets': [('yes', True, 'quiet', True)],
	'text': {
		'desc': {
			'tx':     f'Create, sign and send an {gc.proj_name} transaction',
			'swaptx': f'Create, sign and send a DEX swap transaction from one {gc.proj_name} tracking wallet to another',
		}[target],
		'usage':   '[opts] {u_args} [addr file ...] [seed source ...]',
		'options': """
			-- -h, --help             Print this help message
			-- --, --longhelp         Print help message for long (global) options
			r- -A, --fee-adjust=    f Adjust transaction fee by factor ‘f’ (see below)
			-- -b, --brain-params=l,p Use seed length 'l' and hash preset 'p' for
			+                         brainwallet input
			-- -B, --no-blank         Don't blank screen before displaying {a_info}
			-- -c, --comment-file=  f Source the transaction's comment from file 'f'
			b- -C, --fee-estimate-confs=c Desired number of confirmations for fee estimation
			+                         (default: {cfg.fee_estimate_confs})
			-- -d, --outdir=        d Specify an alternate directory 'd' for output
			e- -D, --contract-data= D Path to file containing hex-encoded contract data
			-- -e, --echo-passphrase  Print passphrase to screen when typing it
			b- -E, --fee-estimate-mode=M Specify the network fee estimate mode.  Choices:
			+                         {fe_all}.  Default: {fe_dfl!r}
			r- -f, --fee=           f Transaction fee, as a decimal {cu} amount or as
			+                         {fu} (an integer followed by {fl}).
			+                         See FEE SPECIFICATION below.  If omitted, fee will be
			+                         calculated using network fee estimation.
			et -g, --gas=N            Set the gas limit (see GAS LIMIT below)
			-s -g, --gas=N            Set the gas limit for Ethereum (see GAS LIMIT below)
			-s -G, --router-gas=N     Set the gas limit for the Ethereum router contract
			+                         (integer).  When unset, a hardcoded default will be
			+                         used.  Applicable only for swaps from token assets.
			-- -H, --hidden-incog-input-params=f,o  Read hidden incognito data from file
			+                        'f' at offset 'o' (comma-separated)
			-- -i, --in-fmt=        f Input is from wallet format 'f' (see FMT CODES below)
			-- -I, --inputs=        i Specify transaction inputs (comma-separated list of
			+                         MMGen IDs or coin addresses).  Note that ALL unspent
			+                         outputs associated with each address will be included.
			-- -l, --seed-len=      l Specify wallet seed length of 'l' bits. This option
			+                         is required only for brainwallet and incognito inputs
			+                         with non-standard (< {dsl}-bit) seed lengths.
			-- -k, --keys-from-file=f Provide additional keys for non-{pnm} addresses
			-- -K, --keygen-backend=n Use backend 'n' for public key generation.  Options
			+                         for {coin_id}: {kgs}
			-s -l, --trade-limit=L    Minimum swap amount, as either percentage or absolute
			+                         coin amount (see TRADE LIMIT below)
			bt -l, --locktime=      t Lock time (block height or unix seconds) (default: 0)
			b- -L, --autochg-ignore-labels Ignore labels when autoselecting change addresses
			-- -m, --minconf=n        Minimum number of confirmations required to spend
			+                         outputs (default: 1)
			-- -M, --mmgen-keys-from-file=f Provide keys for {pnm} addresses in a key-
			+                         address file (output of '{pnl}-keygen'). Permits
			+                         online signing without an {pnm} seed source. The
			+                         key-address file is also used to verify {pnm}-to-{cu}
			+                         mappings, so the user should record its checksum.
			e- -n, --tx-proxy=P       Send transaction via public TX proxy ‘P’ (supported
			+                         proxies: {tx_proxies}).  This is done via a publicly
			+                         accessible web page, so no API key or registration is
			+                         required.
			-- -O, --old-incog-fmt    Specify old-format incognito input
			-- -p, --hash-preset=   p Use the scrypt hash parameters defined by preset 'p'
			+                         for password hashing (default: '{gc.dfl_hash_preset}')
			-- -P, --passwd-file=   f Get {pnm} wallet passphrase from file 'f'
			-- -q, --quiet            Suppress warnings; overwrite files without prompting
			-s -r, --stream-interval=N Set block interval for streaming swap (default: {si})
			bt -R, --no-rbf           Make transaction non-replaceable (non-replace-by-fee
			+                         according to BIP 125)
			-s -s, --swap-proto       Swap protocol to use (Default: {x_dfl},
			+                         Choices: {x_all})
			-- -T, --txhex-idx=N      Send only part ‘N’ of a multi-part transaction.
			+                         Indexing begins with one.
			-- -u, --subseeds=      n The number of subseed pairs to scan for (default: {ss},
			+                         maximum: {ss_max}). Only the default or first supplied
			+                         wallet is scanned for subseeds.
			-- -v, --verbose          Produce more verbose output
			b- -V, --vsize-adj=     f Adjust transaction's estimated vsize by factor 'f'
			e- -w, --wait             Wait for transaction confirmation
			rs -x, --proxy=P          Fetch the swap quote via SOCKS5h proxy ‘P’ (host:port).
			+                         Use special value ‘env’ to honor *_PROXY environment
			+                         vars instead.
			X- -x, --proxy=P          Connect to remote server(s) via SOCKS5h proxy ‘P’
			+                         (host:port).  Use special value ‘env’ to honor *_PROXY
			+                         environment vars instead.
			e- -X, --cached-balances  Use cached balances
			-- -y, --yes              Answer 'yes' to prompts, suppress non-essential output
			-- -z, --show-hash-presets Show information on available hash presets
		""",
		'notes': """
{c}
{n_at}

{g}{F}
                                 SIGNING NOTES
{s}
Seed source files must have the canonical extensions listed in the 'FileExt'
column below:

{f}

{x}"""
	},
	'code': {
		'usage': lambda cfg, proto, help_notes, s: s.format(
			u_args  = help_notes(f'{target}create_args')),
		'options': lambda cfg, proto, help_notes, s: s.format(
			gc      = gc,
			cfg     = cfg,
			cu      = proto.coin,
			pnm     = gc.proj_name,
			pnl     = gc.proj_name.lower(),
			a_info  = help_notes('account_info_desc'),
			kgs     = help_notes('keygen_backends'),
			coin_id = help_notes('coin_id'),
			fu      = help_notes('rel_fee_desc'),
			fl      = help_notes('fee_spec_letters', use_quotes=True),
			dsl     = help_notes('dfl_seed_len'),
			ss      = help_notes('dfl_subseeds'),
			si      = help_notes('stream_interval'),
			tx_proxies = help_notes('tx_proxies'),
			ss_max  = SubSeedIdxRange.max_idx,
			fe_all  = fmt_list(cfg._autoset_opts['fee_estimate_mode'].choices, fmt='no_spc'),
			fe_dfl  = cfg._autoset_opts['fee_estimate_mode'].choices[0],
			x_all   = fmt_list(cfg._autoset_opts['swap_proto'].choices, fmt='no_spc'),
			x_dfl   = cfg._autoset_opts['swap_proto'].choices[0]),
		'notes': lambda cfg, help_mod, help_notes, s: s.format(
			c       = help_mod(f'{target}create'),
			g       = help_notes('gas_limit', target),
			F       = help_notes('fee'),
			n_at    = help_notes('address_types'),
			f       = help_notes('fmt_codes'),
			s       = help_mod('txsign'),
			x       = help_mod(f'{target}create_examples'))
	}
}

cfg = Config(opts_data=opts_data)

from .tx import NewTX, SentTX
from .tx.keys import TxKeys, pop_seedfiles

seedfiles = pop_seedfiles(cfg)

async def main():

	if target == 'swaptx':
		from .tx.new_swap import get_send_proto
		proto = get_send_proto(cfg)
	else:
		proto = cfg._proto

	tx1 = NewTX(cfg=cfg, proto=proto, target=target)

	tx2 = await tx1.create(
		cmd_args = cfg._args,
		locktime = int(cfg.locktime or 0),
		caller   = 'txdo')

	tx3 = await tx2.sign(TxKeys(cfg, tx2, seedfiles=seedfiles).keys)

	if tx3:
		tx3.file.write(ask_write=False)
		tx4 = await SentTX(cfg=cfg, data=tx3.__dict__)
		await tx4.send(cfg, asi=None)
	else:
		die(2, 'Transaction could not be signed')

async_run(main())
